"""
Routines for reading ODIM_H5 files.

"""

import datetime

import numpy as np

try:
    import h5py

    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency
from ..io.common import _test_arguments, make_time_unit_str

ODIM_H5_FIELD_NAMES = {
    "TH": "total_power_horizontal",  # uncorrected reflectivity, horizontal
    "TV": "total_power_vertical",  # uncorrected reflectivity, vertical
    "DBZH": "reflectivity_horizontal",  # corrected reflectivity, horizontal
    "DBZV": "reflectivity_vertical",  # corrected reflectivity, vertical
    "ZDR": "differential_reflectivity",  # differential reflectivity
    "RHOHV": "cross_correlation_ratio",
    "LDR": "linear_polarization_ratio",
    "PHIDP": "differential_phase",
    "KDP": "specific_differential_phase",
    "SQI": "normalized_coherent_power",
    "SNR": "signal_to_noise_ratio",
    "VRAD": "velocity",  # radial velocity, marked for deprecation in ODIM HDF5 2.2
    "VRADH": "velocity_horizontal",  # radial velocity, horizontal polarisation
    "VRADV": "velocity_vertical",  # radial velocity, vertical polarisation
    "WRAD": "spectrum_width",
    "QIND": "quality_index",
}


def read_odim_h5(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    include_datasets=None,
    exclude_datasets=None,
    **kwargs
):
    """
    Read a ODIM_H5 file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    include_datasets : list or None, optional
        List of datasets to include from the HDF5 file, given
        as ["dataset1", "dataset2", ...]. Set to None to include all datasets
        not specified by exclude_datasets.
    exclude_datasets : list or None, optional
        List of datasets to exclude from the HDF5 file, given
        as ["dataset1", "dataset2", ...]. Set to None to include all datasets
        specified by include_datasets.

    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_odim_h5 but is not installed"
        )

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = ODIM_H5_FIELD_NAMES
    filemetadata = FileMetadata(
        "odim_h5",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # open the file
    with h5py.File(filename, "r") as hfile:
        odim_object = _to_str(hfile["what"].attrs["object"])
        if odim_object not in ["PVOL", "SCAN", "ELEV", "AZIM"]:
            raise NotImplementedError("object: %s not implemented." % (odim_object))

        # determine the number of sweeps by the number of groups which
        # begin with dataset
        if include_datasets is not None:
            datasets = [
                k for k in hfile if k.startswith("dataset") and k in include_datasets
            ]
        elif exclude_datasets is not None:
            datasets = [
                k
                for k in hfile
                if k.startswith("dataset") and k not in exclude_datasets
            ]
        else:
            datasets = [k for k in hfile if k.startswith("dataset")]
        datasets.sort(key=lambda x: int(x[7:]))
        nsweeps = len(datasets)

        # latitude, longitude and altitude
        latitude = filemetadata("latitude")
        longitude = filemetadata("longitude")
        altitude = filemetadata("altitude")

        h_where = hfile["where"].attrs
        latitude["data"] = np.array([h_where["lat"]], dtype="float64")
        longitude["data"] = np.array([h_where["lon"]], dtype="float64")
        altitude["data"] = np.array([h_where["height"]], dtype="float64")

        # metadata
        metadata = filemetadata("metadata")
        metadata["source"] = _to_str(hfile["what"].attrs["source"])
        metadata["original_container"] = "odim_h5"
        metadata["odim_conventions"] = _to_str(hfile.attrs["Conventions"])

        h_what = hfile["what"].attrs
        metadata["version"] = _to_str(h_what["version"])
        metadata["source"] = _to_str(h_what["source"])

        try:
            ds1_how = hfile[datasets[0]]["how"].attrs
        except KeyError:
            # if no how group exists mock it with an empty dictionary
            ds1_how = {}
        if "system" in ds1_how:
            metadata["system"] = ds1_how["system"]
        if "software" in ds1_how:
            metadata["software"] = ds1_how["software"]
        if "sw_version" in ds1_how:
            metadata["sw_version"] = ds1_how["sw_version"]

        # sweep_start_ray_index, sweep_end_ray_index
        sweep_start_ray_index = filemetadata("sweep_start_ray_index")
        sweep_end_ray_index = filemetadata("sweep_end_ray_index")

        if odim_object in ["AZIM", "SCAN", "PVOL"]:
            rays_per_sweep = [int(hfile[d]["where"].attrs["nrays"]) for d in datasets]
        elif odim_object == "ELEV":
            rays_per_sweep = [
                int(hfile[d]["where"].attrs["angles"].size) for d in datasets
            ]
        total_rays = sum(rays_per_sweep)
        ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype("int32")
        seri = np.cumsum(rays_per_sweep).astype("int32") - 1
        sweep_start_ray_index["data"] = ssri
        sweep_end_ray_index["data"] = seri

        # sweep_number
        sweep_number = filemetadata("sweep_number")
        sweep_number["data"] = np.arange(nsweeps, dtype="int32")

        # sweep_mode
        sweep_mode = filemetadata("sweep_mode")
        sweep_mode["data"] = np.array(nsweeps * ["azimuth_surveillance"])

        # scan_type
        if odim_object == "ELEV":
            scan_type = "rhi"
        else:
            scan_type = "ppi"

        # fixed_angle
        fixed_angle = filemetadata("fixed_angle")
        if odim_object == "ELEV":
            sweep_el = [hfile[d]["where"].attrs["az_angle"] for d in datasets]
        else:
            sweep_el = [hfile[d]["where"].attrs["elangle"] for d in datasets]
        fixed_angle["data"] = np.array(sweep_el, dtype="float32")

        # elevation
        elevation = filemetadata("elevation")
        if "elangles" in ds1_how:
            edata = np.empty(total_rays, dtype="float32")
            for d, start, stop in zip(datasets, ssri, seri):
                edata[start : stop + 1] = hfile[d]["how"].attrs["elangles"][:]
            elevation["data"] = edata
        elif odim_object == "ELEV":
            edata = np.empty(total_rays, dtype="float32")
            for d, start, stop in zip(datasets, ssri, seri):
                edata[start : stop + 1] = hfile[d]["where"].attrs["angles"][:]
            elevation["data"] = edata
        else:
            elevation["data"] = np.repeat(sweep_el, rays_per_sweep)

        # range
        _range = filemetadata("range")
        if "rstart" in hfile["dataset1/where"].attrs:
            # derive range from rstart and rscale attributes if available

            # check that the gate spacing is constant between sweeps
            rstart = [hfile[d]["where"].attrs["rstart"] for d in datasets]
            if any(rstart != rstart[0]):
                raise ValueError("range start changes between sweeps")
            rscale = [hfile[d]["where"].attrs["rscale"] for d in datasets]
            if any(rscale != rscale[0]):
                raise ValueError("range scale changes between sweeps")
            all_sweeps_nbins = [hfile[d]["where"].attrs["nbins"] for d in datasets]
            # check for max range off all sweeps
            max_nbins = max(all_sweeps_nbins)

            if isinstance(max_nbins, np.ndarray):
                max_nbins = max_nbins[0]
            else:
                max_nbins = max(all_sweeps_nbins)

            rscenter = 1e3 * rstart[0] + rscale[0] / 2
            _range["data"] = np.arange(
                rscenter, rscenter + max_nbins * rscale[0], rscale[0], dtype="float32"
            )
            _range["meters_to_center_of_first_gate"] = rstart[0] * 1000.0
            _range["meters_between_gates"] = float(rscale[0])
        else:
            # if not defined use range attribute which defines the maximum range
            # in km. There is no information on the starting location of the
            # range bins so we assume this to be 0.
            # This most often occurs in RHI files, which technically do not meet
            # the ODIM 2.2 specs. Section 7.4 requires that these files include
            # the where/rstart, where/rscale and where/nbins attributes.
            max_range = [hfile[d]["where"].attrs["range"] for d in datasets]
            if any(max_range != max_range[0]):
                raise ValueError("maximum range changes between sweeps")
            # nbins is required
            nbins = hfile["dataset1/data1/data"].shape[1]
            _range["data"] = np.linspace(0, max_range[0] * 1000.0, nbins).astype(
                "float32"
            )
            _range["meters_to_center_of_first_gate"] = 0
            _range["meters_between_gates"] = max_range[0] * 1000.0 / nbins

        # azimuth
        azimuth = filemetadata("azimuth")
        az_data = np.ones((total_rays,), dtype="float32")
        for dset, start, stop in zip(datasets, ssri, seri):
            if odim_object == "ELEV":
                # all azimuth angles are the sweep azimuth angle
                sweep_az = hfile[dset]["where"].attrs["az_angle"]
            elif odim_object == "AZIM":
                # Sector azimuths are specified in the startaz and stopaz
                # attribute of dataset/where.
                # Assume that the azimuth angles do not pass through 0/360 deg.
                startaz = hfile[dset]["where"].attrs["startaz"]
                stopaz = hfile[dset]["where"].attrs["stopaz"]
                nrays = stop - start + 1
                sweep_az = np.linspace(startaz, stopaz, nrays, endpoint=True)
            elif ("startazA" in ds1_how) and ("stopazA" in ds1_how):
                # average between start and stop azimuth angles
                startaz = hfile[dset]["how"].attrs["startazA"]
                stopaz = hfile[dset]["how"].attrs["stopazA"]
                sweep_az = np.angle(
                    (
                        np.exp(1.0j * np.deg2rad(startaz))
                        + np.exp(1.0j * np.deg2rad(stopaz))
                    )
                    / 2.0,
                    deg=True,
                )
            else:
                # according to section 5.1 the first ray points to the
                # northernmost direction and proceeds clockwise for a complete
                # 360 rotation.
                try:
                    astart = hfile[dset]["how"].attrs["astart"]
                except KeyError:
                    astart = 0.0
                nrays = hfile[dset]["where"].attrs["nrays"]
                da = 360.0 / nrays
                sweep_az = np.arange(astart + da / 2.0, 360.0, da, dtype="float32")
            az_data[start : stop + 1] = sweep_az
        azimuth["data"] = az_data

        # time
        _time = filemetadata("time")
        if ("startazT" in ds1_how) and ("stopazT" in ds1_how):
            # average between startazT and stopazT
            t_data = np.empty((total_rays,), dtype="float32")
            for dset, start, stop in zip(datasets, ssri, seri):
                t_start = hfile[dset]["how"].attrs["startazT"]
                t_stop = hfile[dset]["how"].attrs["stopazT"]
                t_data[start : stop + 1] = (t_start + t_stop) / 2
            start_epoch = t_data.min()
            start_time = datetime.datetime.utcfromtimestamp(start_epoch)
            _time["units"] = make_time_unit_str(start_time)
            _time["data"] = t_data - start_epoch
        else:
            t_data = np.empty((total_rays,), dtype="int32")
            # interpolate between each sweep starting and ending time
            for dset, start, stop in zip(datasets, ssri, seri):
                dset_what = hfile[dset]["what"].attrs
                start_str = _to_str(dset_what["startdate"] + dset_what["starttime"])
                end_str = _to_str(dset_what["enddate"] + dset_what["endtime"])
                start_dt = datetime.datetime.strptime(start_str, "%Y%m%d%H%M%S")
                end_dt = datetime.datetime.strptime(end_str, "%Y%m%d%H%M%S")

                time_delta = end_dt - start_dt
                delta_seconds = time_delta.seconds + time_delta.days * 3600 * 24
                rays = stop - start + 1
                sweep_start_epoch = (
                    start_dt - datetime.datetime(1970, 1, 1)
                ).total_seconds()
                t_data[start : stop + 1] = sweep_start_epoch + np.linspace(
                    0, delta_seconds, rays
                )
            start_epoch = t_data.min()
            start_time = datetime.datetime.utcfromtimestamp(start_epoch)
            _time["units"] = make_time_unit_str(start_time)
            _time["data"] = (t_data - start_epoch).astype("float32")

        # fields
        fields = {}
        h_field_keys = [k for k in hfile["dataset1"] if k.startswith("data")]
        odim_fields = [
            hfile["dataset1"][d]["what"].attrs["quantity"] for d in h_field_keys
        ]
        for odim_field, h_field_key in zip(odim_fields, h_field_keys):
            field_name = filemetadata.get_field_name(_to_str(odim_field))
            if field_name is None:
                continue
            fdata = np.ma.zeros((total_rays, max_nbins), dtype="float32")
            start = 0
            # loop on the sweeps, copy data into correct location in data array
            for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
                try:
                    sweep_data = _get_odim_h5_sweep_data(hfile[dset][h_field_key])
                except KeyError:
                    sweep_data = np.zeros((rays_in_sweep, max_nbins)) + np.NaN
                sweep_nbins = sweep_data.shape[1]
                fdata[start : start + rays_in_sweep, :sweep_nbins] = sweep_data[:]
                # set data to NaN if its beyond the range of this sweep
                fdata[start : start + rays_in_sweep, sweep_nbins:max_nbins] = np.nan
                start += rays_in_sweep
            # create field dictionary
            field_dic = filemetadata(field_name)
            field_dic["data"] = fdata
            field_dic["_FillValue"] = get_fillvalue()
            fields[field_name] = field_dic

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )


def _to_str(text):
    """Convert bytes to str if necessary."""
    if hasattr(text, "decode"):
        return text.decode("utf-8")
    else:
        return text


def _get_odim_h5_sweep_data(group):
    """Get ODIM_H5 sweet data from an HDF5 group."""

    # mask raw data
    what = group["what"]
    raw_data = group["data"][:]

    if "nodata" in what.attrs:
        nodata = what.attrs.get("nodata")
        data = np.ma.masked_equal(raw_data, nodata)
    else:
        data = np.ma.masked_array(raw_data)
    if "undetect" in what.attrs:
        undetect = what.attrs.get("undetect")
        data[data == undetect] = np.ma.masked

    offset = 0.0
    gain = 1.0
    if "offset" in what.attrs:
        offset = what.attrs.get("offset")
    if "gain" in what.attrs:
        gain = what.attrs.get("gain")
    return data * gain + offset



"""
=======================================
Convective-Stratiform classification
=======================================
This example shows how to use the updated convective stratiform classifcation algorithm. We show 3 examples,
a summer convective example, an example from Hurricane Ian, and an example from a winter storm.
"""

print(__doc__)

# Author: Laura Tomkins (lmtomkin@ncsu.edu)
# License: BSD 3 clause


import cartopy.crs as ccrs
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np

import pyart

######################################
# How the algorithm works
# ----------
# This first section describes how the convective-stratiform algorithm works (see references for full details). This
# algorithm is a feature detection algorithm and classifies fields as "convective" or "stratiform". The algorithm is
# designed to detect features in a reflectivity field but can also detect features in fields such as rain rate or
# snow rate. In this section we describe the steps of the convective stratiform algorithm and the variables used in
# the function.
# The first step of the algorithm calculates a background average of the field with a circular footprint using a radius
# provided by ``bkg_rad_km``. A larger radius will yield a smoother field. The radius needs to be at least double the
# grid spacing, but we recommend at least three times the grid spacing. If using reflectivity, ``dB_averaging`` should be set
# to True to convert reflectivity to linear Z before averaging, False for rescaled fields such as rain or snow rate.
# ``calc_thres`` determines the minimum fraction of a circle that is considered in the background average calculation
# (default is 0.75, so the points along the edges where there is less than 75% of a full circle of data,
# the algorithm is not run).
# Once the background average has been calculated, the original field is compared to the background average.  In
# order for points to be considered "convective cores" they must exceed the background value by a certain value or
# simply be greater than the ``always_core_thres``. This value is determined by either a cosine scheme, or a scalar
# value (i.e. the reflectivity value must be X times the background value (multiplier; ``use_addition=False``),
# or  X greater than the background value (``use_addition=True``) where X is the ``scalar_diff``).
# ``use_cosine`` determines if a cosine scheme or scalar scheme is to be used. If ``use_cosine`` is True,
# then the ``max_diff`` and ``zero_diff_cos_val`` come into use. These values define the cosine scheme that is used  to
# determine the minimum difference between the background and reflectivity value in order for a core to be
# identified. ``max_diff`` is the maximum difference between the field and the background for a core to be identified,
# or where the cosine function crosses the y-axis. The ``zero_diff_cos_val`` is where the difference between the  field
# and the background is zero, or where the cosine function crosses the x-axis. Note, if
# ``always_core_thres`` < ``zero_diff_cos_val``, ``zero_diff_cos_val`` only helps define the shape of the cosine curve and
# all values greater than ``always_core_thres`` will be considered a convective core. If
# ``always_core_thres`` > ``zero_diff_cos_val`` then all values greater than ``zero_diff_cos_val`` will be considered a
# convective core. We plot some examples of the schemes below:

# plot cosine scheme
pyart.graph.plot_convstrat_scheme(
    always_core_thres=30, use_cosine=True, max_diff=5, zero_diff_cos_val=45
)
# when zero_diff_cos_val is greater than always_core_thres, the difference becomes zero at the zero_diff_cos_val
pyart.graph.plot_convstrat_scheme(
    always_core_thres=55, use_cosine=True, max_diff=5, zero_diff_cos_val=45
)
# alternatively, we can use a simpler scalar difference instead of a cosine scheme
pyart.graph.plot_convstrat_scheme(
    always_core_thres=40,
    use_cosine=False,
    max_diff=None,
    zero_diff_cos_val=None,
    use_addition=True,
    scalar_diff=2,
)
# if you are interested in picking up weak features, you can also use the scalar difference as a multiplier instead,
# so very weak features do not have to be that different from the background to be classified as convective.
pyart.graph.plot_convstrat_scheme(
    always_core_thres=40,
    use_cosine=False,
    max_diff=None,
    zero_diff_cos_val=None,
    use_addition=False,
    scalar_diff=2,
)

######################################
# Once the cores are identified, there is an option to remove speckles (``remove_small_objects``) smaller than a  given
# size (``min_km2_size``).
# After the convective cores are identified, We then incorporate convective radii using
# ``val_for_max_conv_rad`` and ``max_conv_rad_km``. The convective radii act as a dilation and are used to classify
# additional points around the cores as convective that may not have been identified previously.  The
# ``val_for_max_conv_rad`` is the value where the maximum convective radius is applied and the ``max_conv_rad_km`` is the
# maximum convective radius. Values less than the ``val_for_max_conv_rad`` are assigned a convective radius using a step
# function.
# Finally, the points are classified as NOSFCECHO (threshold set with ``min_dBZ_used``; 0), WEAKECHO (threshold set with
# ``weak_echo_thres``; 3), SF (stratiform; 1), CONV (convective; 2).

######################################
# Examples
# ----------
# **Classification of summer convective example**
#
# Our first example classifies echo from a summer convective event. We use a cosine scheme to classify the convective
# points.

# Now let's do a classification with our parameters
# read in file
filename = pyart.testing.get_test_data("swx_20120520_0641.nc")
radar = pyart.io.read(filename)

# extract the lowest sweep
radar = radar.extract_sweeps([0])

# interpolate to grid
grid = pyart.map.grid_from_radars(
    (radar,),
    grid_shape=(1, 201, 201),
    grid_limits=((0, 10000), (-50000.0, 50000.0), (-50000.0, 50000.0)),
    fields=["reflectivity_horizontal"],
)

# get dx dy
dx = grid.x["data"][1] - grid.x["data"][0]
dy = grid.y["data"][1] - grid.y["data"][0]

# convective stratiform classification
convsf_dict = pyart.retrieve.conv_strat_yuter(
    grid,
    dx,
    dy,
    refl_field="reflectivity_horizontal",
    always_core_thres=40,
    bkg_rad_km=20,
    use_cosine=True,
    max_diff=5,
    zero_diff_cos_val=55,
    weak_echo_thres=10,
    max_conv_rad_km=2,
)

# add to grid object
# mask zero values (no surface echo)
convsf_masked = np.ma.masked_equal(convsf_dict["convsf"]["data"], 0)
# mask 3 values (weak echo)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
# add dimension to array to add to grid object
convsf_dict["convsf"]["data"] = convsf_masked[None, :, :]
# add field
grid.add_field("convsf", convsf_dict["convsf"], replace_existing=True)

# create plot using GridMapDisplay
# plot variables
display = pyart.graph.GridMapDisplay(grid)
magma_r_cmap = plt.get_cmap("magma_r")
ref_cmap = mcolors.LinearSegmentedColormap.from_list(
    "ref_cmap", magma_r_cmap(np.linspace(0, 0.9, magma_r_cmap.N))
)
projection = ccrs.AlbersEqualArea(
    central_latitude=radar.latitude["data"][0],
    central_longitude=radar.longitude["data"][0],
)
######################################
# Note how the convective stratiform field has less data around the edges compared to the reflectivity
# field. This is because the function is designed to only compute the background radius where the footprint has 75%
# of data, so along the edges there is not enough data to fulfill this requirement. The footprint percentage can be
# changed using the variable ``calc_thres``.

# plot
plt.figure(figsize=(10, 4))
ax1 = plt.subplot(1, 2, 1, projection=projection)
display.plot_grid(
    "reflectivity_horizontal",
    vmin=5,
    vmax=45,
    cmap=ref_cmap,
    transform=ccrs.PlateCarree(),
    ax=ax1,
)
ax2 = plt.subplot(1, 2, 2, projection=projection)
display.plot_grid(
    "convsf",
    vmin=0,
    vmax=2,
    cmap=plt.get_cmap("viridis", 3),
    ax=ax2,
    transform=ccrs.PlateCarree(),
    ticks=[1 / 3, 1, 5 / 3],
    ticklabs=["", "Stratiform", "Convective"],
)
plt.show()


######################################
# In addition to the default convective-stratiform classification, the function also returns an underestimate
# (convsf_under) and an overestimate (convsf_over) to take into consideration the uncertainty when choosing
# classification parameters. The under and overestimate use the same parameters, but vary the input field by a
# certain value (default is 5 dBZ, can be changed with ``estimate_offset``). The estimation can be turned off (
# ``estimate_flag=False``), but we recommend keeping it turned on.

# mask weak echo and no surface echo
convsf_masked = np.ma.masked_equal(convsf_dict["convsf"]["data"], 0)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
convsf_dict["convsf"]["data"] = convsf_masked
# underest.
convsf_masked = np.ma.masked_equal(convsf_dict["convsf_under"]["data"], 0)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
convsf_dict["convsf_under"]["data"] = convsf_masked
# overest.
convsf_masked = np.ma.masked_equal(convsf_dict["convsf_over"]["data"], 0)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
convsf_dict["convsf_over"]["data"] = convsf_masked

# Plot each estimation
plt.figure(figsize=(10, 4))
ax1 = plt.subplot(131)
ax1.pcolormesh(
    convsf_dict["convsf"]["data"][0, :, :],
    vmin=0,
    vmax=2,
    cmap=plt.get_cmap("viridis", 3),
)
ax1.set_title("Best estimate")
ax1.set_aspect("equal")
ax2 = plt.subplot(132)
ax2.pcolormesh(
    convsf_dict["convsf_under"]["data"], vmin=0, vmax=2, cmap=plt.get_cmap("viridis", 3)
)
ax2.set_title("Underestimate")
ax2.set_aspect("equal")
ax3 = plt.subplot(133)
ax3.pcolormesh(
    convsf_dict["convsf_over"]["data"], vmin=0, vmax=2, cmap=plt.get_cmap("viridis", 3)
)
ax3.set_title("Overestimate")
ax3.set_aspect("equal")
plt.show()

######################################
# **Tropical example**
#
# Let's get a NEXRAD file from Hurricane Ian

# Read in file
nexrad_file = "s3://noaa-nexrad-level2/2022/09/28/KTBW/KTBW20220928_190142_V06"
radar = pyart.io.read_nexrad_archive(nexrad_file)

# extract the lowest sweep
radar = radar.extract_sweeps([0])

# interpolate to grid
grid = pyart.map.grid_from_radars(
    (radar,),
    grid_shape=(1, 201, 201),
    grid_limits=((0, 10000), (-200000.0, 200000.0), (-200000.0, 200000.0)),
    fields=["reflectivity"],
)

# get dx dy
dx = grid.x["data"][1] - grid.x["data"][0]
dy = grid.y["data"][1] - grid.y["data"][0]

# convective stratiform classification
convsf_dict = pyart.retrieve.conv_strat_yuter(
    grid,
    dx,
    dy,
    refl_field="reflectivity",
    always_core_thres=40,
    bkg_rad_km=20,
    use_cosine=True,
    max_diff=3,
    zero_diff_cos_val=55,
    weak_echo_thres=5,
    max_conv_rad_km=2,
    estimate_flag=False,
)

# add to grid object
# mask zero values (no surface echo)
convsf_masked = np.ma.masked_equal(convsf_dict["convsf"]["data"], 0)
# mask 3 values (weak echo)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
# add dimension to array to add to grid object
convsf_dict["convsf"]["data"] = convsf_masked[None, :, :]
# add field
grid.add_field("convsf", convsf_dict["convsf"], replace_existing=True)

# create plot using GridMapDisplay
# plot variables
display = pyart.graph.GridMapDisplay(grid)
magma_r_cmap = plt.get_cmap("magma_r")
ref_cmap = mcolors.LinearSegmentedColormap.from_list(
    "ref_cmap", magma_r_cmap(np.linspace(0, 0.9, magma_r_cmap.N))
)
projection = ccrs.AlbersEqualArea(
    central_latitude=radar.latitude["data"][0],
    central_longitude=radar.longitude["data"][0],
)
# plot
plt.figure(figsize=(10, 4))
ax1 = plt.subplot(1, 2, 1, projection=projection)
display.plot_grid(
    "reflectivity",
    vmin=5,
    vmax=45,
    cmap=ref_cmap,
    transform=ccrs.PlateCarree(),
    ax=ax1,
    axislabels_flag=False,
)
ax2 = plt.subplot(1, 2, 2, projection=projection)
display.plot_grid(
    "convsf",
    vmin=0,
    vmax=2,
    cmap=plt.get_cmap("viridis", 3),
    axislabels_flag=False,
    transform=ccrs.PlateCarree(),
    ticks=[1 / 3, 1, 5 / 3],
    ticklabs=["", "Stratiform", "Convective"],
    ax=ax2,
)
plt.show()

######################################
# **Winter storm example with image muting**
#
# Here is a final example of the convective stratiform classification using an example from a winter storm. Before
# doing the classification, we image mute the reflectivity to remove regions with melting or mixed precipitation. We
# then rescale the reflectivity to snow rate (Rasumussen et al. 2003). We recommend using a rescaled reflectivity
# to do the classification, but if you do make sure to changed dB_averaging to False because this parameter is  used
# to convert reflectivity to a linear value before averaging (set dB_averaging to True for reflectivity fields in
# dBZ units).
# In this example, note how we change some of the other parameters since we are classifying snow rate instead of
# reflecitivity.

# Read in file
nexrad_file = "s3://noaa-nexrad-level2/2021/02/07/KOKX/KOKX20210207_161413_V06"
radar = pyart.io.read_nexrad_archive(nexrad_file)

# extract the lowest sweep
radar = radar.extract_sweeps([0])

# interpolate to grid
grid = pyart.map.grid_from_radars(
    (radar,),
    grid_shape=(1, 201, 201),
    grid_limits=((0, 10000), (-200000.0, 200000.0), (-200000.0, 200000.0)),
    fields=["reflectivity", "cross_correlation_ratio"],
)

# image mute grid object
grid = pyart.util.image_mute_radar(
    grid, "reflectivity", "cross_correlation_ratio", 0.97, 20
)

# convect non-muted reflectivity to snow rate
nonmuted_ref = grid.fields["nonmuted_reflectivity"]["data"][0, :, :]
nonmuted_ref = np.ma.masked_invalid(nonmuted_ref)

nonmuted_ref_linear = 10 ** (nonmuted_ref / 10)  # mm6/m3
snow_rate = (nonmuted_ref_linear / 57.3) ** (1 / 1.67)  #

# add to grid
snow_rate_dict = {
    "data": snow_rate[None, :, :],
    "standard_name": "snow_rate",
    "long_name": "Snow rate converted from linear reflectivity",
    "units": "mm/hr",
    "valid_min": 0,
    "valid_max": 40500,
}
grid.add_field("snow_rate", snow_rate_dict, replace_existing=True)

# get dx dy
dx = grid.x["data"][1] - grid.x["data"][0]
dy = grid.y["data"][1] - grid.y["data"][0]

# convective stratiform classification
convsf_dict = pyart.retrieve.conv_strat_yuter(
    grid,
    dx,
    dy,
    refl_field="snow_rate",
    dB_averaging=False,
    always_core_thres=4,
    bkg_rad_km=40,
    use_cosine=True,
    max_diff=1.5,
    zero_diff_cos_val=5,
    weak_echo_thres=0,
    min_dBZ_used=0,
    max_conv_rad_km=1,
    estimate_flag=False,
)

# add to grid object
# mask zero values (no surface echo)
convsf_masked = np.ma.masked_equal(convsf_dict["convsf"]["data"], 0)
# mask 3 values (weak echo)
convsf_masked = np.ma.masked_equal(convsf_masked, 3)
# add dimension to array to add to grid object
convsf_dict["convsf"]["data"] = convsf_masked[None, :, :]
# add field
grid.add_field("convsf", convsf_dict["convsf"], replace_existing=True)

# create plot using GridMapDisplay
# plot variables
display = pyart.graph.GridMapDisplay(grid)
magma_r_cmap = plt.get_cmap("magma_r")
ref_cmap = mcolors.LinearSegmentedColormap.from_list(
    "ref_cmap", magma_r_cmap(np.linspace(0, 0.9, magma_r_cmap.N))
)
projection = ccrs.AlbersEqualArea(
    central_latitude=radar.latitude["data"][0],
    central_longitude=radar.longitude["data"][0],
)
# plot
plt.figure(figsize=(10, 4))
ax1 = plt.subplot(1, 2, 1, projection=projection)
display.plot_grid(
    "snow_rate",
    vmin=0,
    vmax=10,
    cmap=plt.get_cmap("viridis"),
    transform=ccrs.PlateCarree(),
    ax=ax1,
    axislabels_flag=False,
)
ax2 = plt.subplot(1, 2, 2, projection=projection)
display.plot_grid(
    "convsf",
    vmin=0,
    vmax=2,
    cmap=plt.get_cmap("viridis", 3),
    axislabels_flag=False,
    transform=ccrs.PlateCarree(),
    ticks=[1 / 3, 1, 5 / 3],
    ticklabs=["", "Stratiform", "Convective"],
    ax=ax2,
)
plt.show()

######################################
# Summary of recommendations and best practices
# ----------
# * Tune your parameters to your specific purpose
# * Use a rescaled field if possible (i.e. linear reflectivity, rain or snow rate)
# * Keep ``estimate_flag=True`` to see uncertainty in classification
#
# References
# ----------
# Steiner, M. R., R. A. Houze Jr., and S. E. Yuter, 1995: Climatological
# Characterization of Three-Dimensional Storm Structure from Operational
# Radar and Rain Gauge Data. J. Appl. Meteor., 34, 1978-2007.
# https://doi.org/10.1175/1520-0450(1995)034<1978:CCOTDS>2.0.CO;2.
#
# Yuter, S. E., and R. A. Houze, Jr., 1997: Measurements of raindrop size
# distributions over the Pacific warm pool and implications for Z-R relations.
# J. Appl. Meteor., 36, 847-867.
# https://doi.org/10.1175/1520-0450(1997)036%3C0847:MORSDO%3E2.0.CO;2
#
# Yuter, S. E., R. A. Houze, Jr., E. A. Smith, T. T. Wilheit, and E. Zipser,
# 2005: Physical characterization of tropical oceanic convection observed in
# KWAJEX. J. Appl. Meteor., 44, 385-415. https://doi.org/10.1175/JAM2206.1
#
# Rasmussen, R., M. Dixon, S. Vasiloff, F. Hage, S. Knight, J. Vivekanandan,
# and M. Xu, 2003: Snow Nowcasting Using a Real-Time Correlation of Radar
# Reflectivity with Snow Gauge Accumulation. J. Appl. Meteorol. Climatol., 42, 20–36.
# https://doi.org/10.1175/1520-0450(2003)042%3C0020:SNUART%3E2.0.CO;2



"""
======================================
Compare Two Radars Using Gatemapper
======================================

Map the reflectivity field of a single radar in Antenna coordinates to
another radar in Antenna coordinates and compare the fields.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov) and Bobby Jackson (rjackson@anl.gov)
# License: BSD 3 clause

import warnings

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

import pyart
from pyart.testing import get_test_data

warnings.filterwarnings("ignore")

######################################
# **Read in the Data**
#
# For this example, we use two XSAPR radars from our test data.

# read in the data from both XSAPR radars
xsapr_sw_file = get_test_data("swx_20120520_0641.nc")
xsapr_se_file = get_test_data("sex_20120520_0641.nc")
radar_sw = pyart.io.read_cfradial(xsapr_sw_file)
radar_se = pyart.io.read_cfradial(xsapr_se_file)

######################################
# **Filter and Configure the GateMapper**
#
# We are interested in mapping the southwestern radar to the
# southeastern radar. Before running our gatemapper, we add a
# filter for only positive reflectivity values.
# We also need to set a distance (meters) and time (seconds)
# between the source and destination gate allowed for an
# adequate match), using the distance_tolerance/time_tolerance variables.

gatefilter = pyart.filters.GateFilter(radar_sw)
gatefilter.exclude_below("reflectivity_horizontal", 20)
gmapper = pyart.map.GateMapper(
    radar_sw,
    radar_se,
    distance_tolerance=500.0,
    time_tolerance=60,
    gatefilter_src=gatefilter,
)
radar_sw_mapped_to_radar_se = gmapper.mapped_radar(["reflectivity_horizontal"])

######################################
# **Plot the Original Data**
#
# Let's take a look at our original fields - notice the difference
# in reflectivity values!

fig = plt.figure(figsize=(16, 6))
ax = plt.subplot(121, projection=ccrs.PlateCarree())

# Plot the southwestern radar
disp1 = pyart.graph.RadarMapDisplay(radar_sw)
disp1.plot_ppi_map(
    "reflectivity_horizontal",
    sweep=1,
    ax=ax,
    vmin=-20,
    vmax=70,
    min_lat=36,
    max_lat=37,
    min_lon=-98,
    max_lon=-97,
    lat_lines=np.arange(36, 37.25, 0.25),
    lon_lines=np.arange(-98, -96.75, 0.25),
)

ax2 = plt.subplot(122, projection=ccrs.PlateCarree())
disp2 = pyart.graph.RadarMapDisplay(radar_se)
disp2.plot_ppi_map(
    "reflectivity_horizontal",
    sweep=1,
    ax=ax2,
    vmin=-20,
    vmax=70,
    min_lat=36,
    max_lat=37,
    min_lon=-98,
    max_lon=-97,
    lat_lines=np.arange(36, 37.25, 0.25),
    lon_lines=np.arange(-98, -96.75, 0.25),
)

######################################
# Now, we can compare our original field from the southwestern radar,
# to the new remapped field - there are similarities...

fig = plt.figure(figsize=(16, 6))
ax = plt.subplot(121, projection=ccrs.PlateCarree())

# Plot the southeastern radar
disp1 = pyart.graph.RadarMapDisplay(radar_se)
disp1.plot_ppi_map(
    "reflectivity_horizontal",
    sweep=1,
    ax=ax,
    vmin=-20,
    vmax=70,
    min_lat=36,
    max_lat=37,
    min_lon=-98,
    max_lon=-97,
    lat_lines=np.arange(36, 37.25, 0.25),
    lon_lines=np.arange(-98, -96.75, 0.25),
)

# Plot the southwestern radar mapped to the southeastern radar
ax2 = plt.subplot(122, projection=ccrs.PlateCarree())
disp2 = pyart.graph.RadarMapDisplay(radar_sw_mapped_to_radar_se)
disp2.plot_ppi_map(
    "reflectivity_horizontal",
    sweep=1,
    ax=ax2,
    vmin=-20,
    vmax=70,
    min_lat=36,
    max_lat=37,
    min_lon=-98,
    max_lon=-97,
    lat_lines=np.arange(36, 37.25, 0.25),
    lon_lines=np.arange(-98, -96.75, 0.25),
)

######################################
# **Calculate and Plot the Difference**
#
# It can be difficult to "eyeball" the difference between these two fields.
# Fortunately, now that our radars match coordinates, we can plot a difference.
# Keep in mind there is a time difference of ~ 1 minute between these plots,
# leading to small difference due to the precipitation moving through the
# domain over the course of that minute.

# Extract the numpy arrays for our reflectivity fields
reflectivity_se_radar = radar_se.fields["reflectivity_horizontal"]["data"]
reflectivity_sw_radar = radar_sw_mapped_to_radar_se.fields["reflectivity_horizontal"][
    "data"
]

# Calculate the difference between the southeastern and southwestern radar
reflectivity_difference = reflectivity_se_radar - reflectivity_sw_radar

# Add a field like this to the radar_se radar object
radar_se.add_field_like(
    "reflectivity_horizontal",
    field_name="reflectivity_bias",
    data=reflectivity_difference,
)

# Setup our figure
fig = plt.figure(figsize=(8, 6))
ax = plt.subplot(111, projection=ccrs.PlateCarree())

# Plot the difference field
disp1 = pyart.graph.RadarMapDisplay(radar_se)
disp1.plot_ppi_map(
    "reflectivity_bias",
    cmap="pyart_balance",
    title="Reflectivity Difference \n XSAPR Southwest - XSPAR Southeast",
    sweep=1,
    ax=ax,
    vmin=-30,
    vmax=30,
    min_lat=36,
    max_lat=37,
    min_lon=-98,
    max_lon=-97,
    lat_lines=np.arange(36, 37.25, 0.25),
    lon_lines=np.arange(-98, -96.75, 0.25),
)


######################################
# **Plot a Histogram for Comparison**
#
# Another way of plotting the comparison here is using
# a 2-dimensional histogram,which is more helpful in this
# case where our scans don't neccessarily match exactly in time.

# Include elevations above the lowest one
incl_gates = np.argwhere(radar_sw_mapped_to_radar_se.elevation["data"] > 1.0)

# Filter the reflectivity fields using the filter created above
refl_se = reflectivity_se_radar[incl_gates, :]
refl_sw = reflectivity_sw_radar[incl_gates, :]

# Make sure not include masked values
values_without_mask = np.logical_and(~refl_se.mask, ~refl_sw.mask)
refl_se = refl_se[values_without_mask]
refl_sw = refl_sw[values_without_mask]

# Set the bins for our histogram
bins = np.arange(-10, 60, 1)

# Create the 2D histogram using the flattened numpy arrays
hist = np.histogram2d(refl_se.flatten(), refl_sw.flatten(), bins=bins)[0]
hist = np.ma.masked_where(hist == 0, hist)

# Setup our figure
fig = plt.figure(figsize=(8, 6))

# Create a 1-1 comparison
x, y = np.meshgrid((bins[:-1] + bins[1:]) / 2.0, (bins[:-1] + bins[1:]) / 2.0)
c = plt.pcolormesh(x, y, np.log10(hist.T), cmap="pyart_HomeyerRainbow")

# Add a colorbar and labels
plt.colorbar(c, label="$log_{10}$ counts")
plt.xlabel("XSAPR Southeast $Z_{H}$ [dBZ]")
plt.ylabel("XSAPR Southwest $Z_{H}$ [dBZ]")

plt.show()



""" Unit Tests for Py-ART's retrieve/echo_class.py module. """

import numpy as np
import pytest
from numpy.testing import assert_allclose

import pyart


def test_steiner_conv_strat_default():
    grid = pyart.testing.make_storm_grid()
    eclass_default = pyart.retrieve.steiner_conv_strat(grid)
    assert np.all(
        eclass_default["data"][25]
        == np.array(
            [
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
            ]
        )
    )


@pytest.mark.parametrize("area_relation", ["small", "medium", "large", "sgp"])
def test_steiner_conv_strat_modify_area(area_relation):
    grid = pyart.testing.make_storm_grid()
    eclass = pyart.retrieve.steiner_conv_strat(grid, area_relation=area_relation)
    assert eclass["data"].min() == 0
    assert eclass["data"].max() == 2


def test_conv_strat__yuter_default():
    grid = pyart.testing.make_storm_grid()
    dict = pyart.retrieve.conv_strat_yuter(grid, bkg_rad_km=50)

    assert "convsf" in dict.keys()
    assert "convsf_under" in dict.keys()
    assert "convsf_over" in dict.keys()
    assert np.all(
        dict["convsf"]["data"][25]
        == np.array(
            [
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
            ]
        )
    )


def test_conv_strat_yuter_noest():
    grid = pyart.testing.make_storm_grid()
    dict = pyart.retrieve.conv_strat_yuter(grid, bkg_rad_km=50, estimate_flag=False)

    assert "convsf" in dict.keys()
    assert "convsf_under" not in dict.keys()
    assert "convsf_over" not in dict.keys()


def test_hydroclass_semisupervised():
    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
    temp_dict = pyart.config.get_metadata("temperature")
    temp = np.empty(radar.fields["reflectivity"]["data"].shape)
    temp[:] = -10.0
    temp_dict["data"] = temp
    radar.add_field("temperature", temp_dict, replace_existing=True)
    radar.fields["specific_differential_phase"] = radar.fields.pop("differential_phase")
    radar_small = radar.extract_sweeps([0])

    mass_centers = pyart.retrieve.echo_class._get_mass_centers(10e9)

    hydro_nofreq = pyart.retrieve.hydroclass_semisupervised(radar_small)
    assert "units" in hydro_nofreq.keys()
    assert "standard_name" in hydro_nofreq.keys()
    assert "long_name" in hydro_nofreq.keys()
    assert "coordinates" in hydro_nofreq.keys()

    assert hydro_nofreq["data"].dtype == "int64"
    assert_allclose(hydro_nofreq["data"][0][0:5], [6, 6, 6, 6, 6])
    assert_allclose(hydro_nofreq["data"][0][-5:], [2, 2, 2, 2, 2])

    radar_small.instrument_parameters["frequency"] = {"data": np.array([5e9])}
    hydro_freq = pyart.retrieve.hydroclass_semisupervised(radar_small)
    assert_allclose(hydro_freq["data"][0][0:5], [6, 6, 6, 6, 6])
    assert_allclose(hydro_freq["data"][0][-5:], [2, 2, 2, 2, 2])

    hydro = pyart.retrieve.hydroclass_semisupervised(
        radar_small, mass_centers=mass_centers
    )
    assert_allclose(hydro["data"][0][0:5], [6, 6, 6, 6, 6])
    assert_allclose(hydro["data"][0][-5:], [2, 2, 2, 2, 2])


def test_data_limits_table():
    dlimits_dict = pyart.retrieve.echo_class._data_limits_table()
    test_dict = {
        "Zh": (60.0, -10.0),
        "ZDR": (5.0, -5.0),
        "KDP": (7.0, -10.0),
        "RhoHV": (-5.23, -50.0),
    }

    assert type(dlimits_dict) == dict
    assert dlimits_dict == test_dict


def test_get_freq():
    freq_band_s = pyart.retrieve.get_freq_band(3e9)
    assert freq_band_s == "S"

    freq_band_c = pyart.retrieve.get_freq_band(6e9)
    assert freq_band_c == "C"

    freq_band_x = pyart.retrieve.get_freq_band(10e9)
    assert freq_band_x == "X"

    freq_band_bad = pyart.retrieve.get_freq_band(10)
    assert freq_band_bad is None


def test_mass_centers_dict():
    mass_centers_dict = pyart.retrieve.echo_class._mass_centers_table()

    assert type(mass_centers_dict)
    assert "C" in mass_centers_dict.keys()
    assert "X" in mass_centers_dict.keys()

    assert_allclose(
        mass_centers_dict["C"][0],
        [1.35829e01, 4.06300e-01, 4.97000e-02, 9.86800e-01, 1.33030e03],
        atol=1e-7,
    )
    assert_allclose(
        mass_centers_dict["C"][8],
        [5.06186e01, -6.49000e-02, 9.46000e-02, 9.90400e-01, 1.17990e03],
        atol=1e-7,
    )

    assert_allclose(
        mass_centers_dict["X"][0],
        [1.9077e01, 4.1390e-01, 9.9000e-03, 9.8410e-01, 1.0617e03],
        atol=1e-7,
    )

    assert_allclose(
        mass_centers_dict["X"][8],
        [4.42216e01, -3.41900e-01, 6.87000e-02, 9.68300e-01, 1.27270e03],
        atol=1e-7,
    )


def test_mass_centers():
    mass_centers_good_freq = pyart.retrieve.echo_class._get_mass_centers(10e9)
    assert_allclose(
        mass_centers_good_freq[0],
        [1.90770e01, 4.13900e-01, 9.90000e-03, 9.84100e-01, 1.06170e03],
        atol=1e-7,
    )

    low_bad_freq = pyart.retrieve.echo_class._get_mass_centers(3e9)
    assert_allclose(
        low_bad_freq[0],
        [1.35829e01, 4.06300e-01, 4.97000e-02, 9.86800e-01, 1.33030e03],
        atol=1e-7,
    )

    high_bad_freq = pyart.retrieve.echo_class._get_mass_centers(13e9)
    assert_allclose(
        high_bad_freq[0],
        [1.90770e01, 4.13900e-01, 9.90000e-03, 9.84100e-01, 1.06170e03],
        atol=1e-7,
    )


def test_assign_to_class():
    rhohv = np.array(([0.2, 0.2, 0.2, 0.2, 0.2], [0.2, 0.2, 0.2, 0.2, 0.2]))
    zdr = np.array(([-7.8, -7.8, -7.8, -7.8, -7.8], [-7.8, -7.8, -7.8, -7.8, -7.8]))
    kdp = np.array(
        ([180.5, 180.5, 180.5, 180.5, 180.5], [180.5, 180.5, 180.5, 180.5, 180.5])
    )
    relh = np.array(([10.0, 10.0, 10.0, 10.0, 10.0], [10.0, 10.0, 10.0, 10.0, 10.0]))
    zh = np.array(
        ([-32.0, -32.0, -32.0, -32.0, -32.0], [-32.0, -32.0, -32.0, -32.0, -32.0])
    )

    mass_centers = pyart.retrieve.echo_class._get_mass_centers(10e9)
    hydroclass, min_dist = pyart.retrieve.echo_class._assign_to_class(
        zh, zdr, kdp, rhohv, relh, mass_centers
    )

    assert_allclose(hydroclass[0], [7, 7, 7, 7, 7], atol=1e-7)
    assert_allclose(
        min_dist[0],
        [227.0343910, 227.0343910, 227.0343910, 227.0343910, 227.0343910],
        atol=1e-7,
    )


def test_standardize():
    kdp = np.array(
        ([180.5, 180.5, 180.5, 180.5, 180.5], [180.5, 180.5, 180.5, 180.5, 180.5])
    )
    field_std_kdp = pyart.retrieve.echo_class._standardize(
        kdp, "KDP", mx=180.0, mn=181.0
    )
    assert_allclose(field_std_kdp[0], [-1.0, -1.0, -1.0, -1.0, -1.0], atol=1e-6)

    relh = np.array(([10.0, 10.0, 10.0, 10.0, 10.0], [10.0, 10.0, 10.0, 10.0, 10.0]))
    field_std_relh = pyart.retrieve.echo_class._standardize(
        relh, "relH", mx=9.0, mn=5.0
    )
    assert_allclose(
        field_std_relh[0], [0.024994, 0.024994, 0.024994, 0.024994, 0.024994], atol=1e-6
    )

    rhohv = np.array(([0.2, 0.2, 0.2, 0.2, 0.2], [0.2, 0.2, 0.2, 0.2, 0.2]))
    field_std_rhohv = pyart.retrieve.echo_class._standardize(
        rhohv, "RhoHV", mx=0.1, mn=1.0
    )
    assert_allclose(field_std_rhohv[0], [-1.0, -1.0, -1.0, -1.0, -1.0], atol=1e-6)

    field_std_no_limits = pyart.retrieve.echo_class._standardize(
        rhohv, "RhoHV", mx=None, mn=None
    )
    assert_allclose(field_std_no_limits[0], [1.0, 1.0, 1.0, 1.0, 1.0], atol=1e-6)

    pytest.raises(ValueError, pyart.retrieve.echo_class._standardize, rhohv, "foo")



""" Unit Tests for Py-ART's _debug_info module. """

import sys
import warnings

import pyart

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def test_debug_info():
    # test to see that something is written when _debug_info is called
    # we don't care what is written, just that something is.
    buf = StringIO()
    pyart._debug_info(buf)
    assert len(buf.getvalue()) > 0


def test_debug_stdout():
    # lack of error is assumed to mean that call succeed
    pyart._debug_info()


# Class for Mocking ImportErrors from
# http://stackoverflow.com/questions/2481511/mocking-importerror-in-python
class DisableModules:
    def __init__(self, modules):
        self.modules = modules

    def find_module(self, fullname):
        if fullname in self.modules:
            raise ImportError("Debug import failure for %s" % fullname)


def test_debug_info_all_disabled():
    modules = [
        "numpy",
        "scipy",
        "matplotlib",
        "netCDF4",
        "cylp",
        "glpk",
        "cvxopt",
        "mpl_toolkits",
        "platform",
    ]
    save_dict = {}
    for module in modules:
        if module in sys.modules:
            save_dict[module] = sys.modules[module]
            sys.modules.pop(module)
    fail_loader = DisableModules(modules)
    sys.meta_path.append(fail_loader)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # test to see that something is written when _debug_info is called
        # we don't care what is written, just that something is.
        buf = StringIO()
        pyart._debug_info(buf)
        assert len(buf.getvalue()) > 0
    # remove the Mocked ImportErrors
    sys.meta_path.remove(fail_loader)
    for module in save_dict.keys():
        sys.modules[module] = save_dict[module]



"""
Classes for specifying what gates are included and excluded from routines.

"""

from .gatefilter import GateFilter  # noqa
from .gatefilter import iso0_based_gate_filter  # noqa
from .gatefilter import moment_and_texture_based_gate_filter  # noqa
from .gatefilter import moment_based_gate_filter  # noqa
from .gatefilter import temp_based_gate_filter  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



"""
Miscellaneous utility functions.

The location and names of these functions within Py-ART may change between
versions without depreciation, use with caution.

"""

from .circular_stats import angular_mean  # noqa
from .circular_stats import angular_mean_deg  # noqa
from .circular_stats import angular_std  # noqa
from .circular_stats import angular_std_deg  # noqa
from .circular_stats import interval_mean  # noqa
from .circular_stats import interval_std  # noqa
from .circular_stats import mean_of_two_angles  # noqa
from .circular_stats import mean_of_two_angles_deg  # noqa
from .columnsect import for_azimuth  # noqa
from .columnsect import get_column_rays  # noqa
from .columnsect import get_field_location  # noqa
from .columnsect import sphere_distance  # noqa
from .datetime_utils import datetime_from_dataset  # noqa
from .datetime_utils import datetime_from_grid  # noqa
from .datetime_utils import datetime_from_radar  # noqa
from .datetime_utils import datetimes_from_dataset  # noqa
from .datetime_utils import datetimes_from_radar  # noqa
from .hildebrand_sekhon import estimate_noise_hs74  # noqa
from .radar_utils import (  # noqa
    image_mute_radar,
    is_vpt,
    join_radar,
    subset_radar,
    to_vpt,
)
from .sigmath import (  # noqa
    angular_texture_2d,
    rolling_window,
    texture,
    texture_along_ray,
)
from .simulated_vel import simulated_vel_from_profile  # noqa
from .xsect import cross_section_ppi, cross_section_rhi  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



""" Unit Tests for Py-ART's uf_write module. """

try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

import datetime
import struct
import warnings

import netCDF4
import numpy as np

import pyart
from pyart.io.uf_write import UF_MISSING_VALUE, UFRayCreator, write_uf
from pyart.io.uffile import UFFile

TEMPLATES_EXTRA = {
    "mandatory_header": {
        "radar_name": b"xsapr-sg",
        "site_name": b"xsapr-sg",
        "generation_year": 15,
        "generation_month": 8,
        "generation_day": 19,
        "generation_facility_name": b"RSLv1.48",
    },
    "optional_header": {
        "project_name": b"TRMMGVUF",
        "tape_name": b"RADAR_UF",
    },
}

FIELD_MAPPING = {
    "DZ": "DZ",
    "VR": "VR",
    "SW": "SW",
    "CZ": "CZ",
    "ZT": "ZT",
    "DR": "DR",
    "ZD": "ZD",
    "RH": "RH",
    "PH": "PH",
    "KD": "KD",
    "SQ": "SQ",
    "HC": "HC",
}


def test_ray_section_by_section():

    ufile = UFFile(pyart.testing.UF_FILE)
    uray = ufile.rays[0]
    ref_ray_buf = uray._buf
    ufile.close()

    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    volume_start = netCDF4.num2date(radar.time["data"][0], radar.time["units"])
    volume_start -= datetime.timedelta(seconds=8)
    nfields = len(radar.fields)
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]
    ufraycreator = UFRayCreator(
        radar,
        FIELD_MAPPING,
        field_write_order,
        volume_start=volume_start,
        templates_extra=TEMPLATES_EXTRA,
    )

    # mandatory header
    ref_man_header = ref_ray_buf[:90]
    tst_man_header = ufraycreator.make_mandatory_header(0)
    assert tst_man_header == ref_man_header

    # optional header
    ref_opt_header = uray._buf[90:118]
    tst_opt_header = ufraycreator.make_optional_header()
    assert tst_opt_header == ref_opt_header

    # data headers
    ref_data_header = uray._buf[118:124]
    tst_data_header = ufraycreator.make_data_header()
    assert tst_data_header == ref_data_header

    # field position info
    ref_field_position = uray._buf[124 : 124 + 4 * nfields]
    tst_field_position = ufraycreator.make_field_position()
    assert tst_field_position == ref_field_position

    # DZ field header
    ref_field_header = uray._buf[172 : 172 + 38]
    ufraycreator.field_header_template["edit_code"] = b"\x00\x00"
    tst_field_header = ufraycreator.make_field_header(106, 0, 100)
    assert tst_field_header == ref_field_header

    # DZ data
    ref_dz_data = uray.field_raw_data[0]
    tst_dz_data = ufraycreator.make_data_array("DZ", 0)
    assert np.array_equal(ref_dz_data, tst_dz_data)

    # DZ data buffer
    ref_dz_data_buf = uray._buf[210:1544]
    assert ref_dz_data_buf == ref_dz_data.tobytes()
    assert ref_dz_data_buf == tst_dz_data.tobytes()

    # VR field header
    ref_field_header = uray._buf[1544 : 1544 + 42]
    ufraycreator.field_header_template["edit_code"] = b"  "
    tst_field_header = ufraycreator.make_field_header(794, 0, 100)
    vel_header = ufraycreator.make_fsi_vel(0, 100)
    assert tst_field_header + vel_header == ref_field_header

    # VR data
    ref_vr_data = uray.field_raw_data[1]
    tst_vr_data = ufraycreator.make_data_array("VR", 0)
    assert np.array_equal(ref_vr_data, tst_vr_data)

    # VR data buffer
    ref_vr_data_buf = uray._buf[1586:2920]
    assert ref_vr_data_buf == ref_vr_data.tobytes()
    assert ref_vr_data_buf == tst_vr_data.tobytes()

    # SW field header
    ref_field_header = uray._buf[2920 : 2920 + 38]
    ufraycreator.field_header_template["edit_code"] = b"  "
    tst_field_header = ufraycreator.make_field_header(1480, 0, 100)
    assert tst_field_header == ref_field_header

    # SW data
    ref_sw_data = uray.field_raw_data[2]
    tst_sw_data = ufraycreator.make_data_array("SW", 0)
    assert np.array_equal(ref_sw_data, tst_sw_data)

    # SW data buffer
    ref_sw_data_buf = uray._buf[2958:4292]
    assert ref_sw_data_buf == ref_sw_data.tobytes()
    assert ref_sw_data_buf == tst_sw_data.tobytes()

    # ZT field header
    ref_field_header = uray._buf[5664 : 5664 + 38]
    ufraycreator.field_header_template["edit_code"] = b"\x00\x00"
    tst_field_header = ufraycreator.make_field_header(2852, 0, 100)
    assert tst_field_header == ref_field_header

    # PH field header
    ref_field_header = uray._buf[11152 : 11152 + 38]
    ufraycreator.field_header_template["edit_code"] = b"  "
    tst_field_header = ufraycreator.make_field_header(5596, 0, 10)
    assert tst_field_header == ref_field_header

    # PH data
    ref_ph_data = uray.field_raw_data[8]
    tst_ph_data = ufraycreator.make_data_array("PH", 0, 10.0)
    assert np.array_equal(ref_ph_data, tst_ph_data)

    # PH data buffer
    ref_ph_data_buf = uray._buf[11152 + 38 : 12524]
    assert ref_ph_data_buf == ref_ph_data.tobytes()
    assert ref_ph_data_buf == tst_ph_data.tobytes()


def test_ray_full():

    ufile = UFFile(pyart.testing.UF_FILE)
    ref_ray = ufile.rays[0]._buf[:]
    ufile.close()

    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    radar.fields["PH"]["_UF_scale_factor"] = 10
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]
    volume_start = netCDF4.num2date(radar.time["data"][0], radar.time["units"])
    volume_start -= datetime.timedelta(seconds=8)
    ufraycreator = UFRayCreator(
        radar,
        FIELD_MAPPING,
        field_write_order,
        volume_start=volume_start,
        templates_extra=TEMPLATES_EXTRA,
    )
    tst_ray = ufraycreator.make_ray(0)
    tst_ray = tst_ray[:204] + b"\x00\x00" + tst_ray[206:]  # DZ edit_code
    tst_ray = tst_ray[:5696] + b"\x00\x00" + tst_ray[5698:]  # ZT edit_code
    assert ref_ray == tst_ray


def test_complete_file():

    with open(pyart.testing.UF_FILE, "rb") as fh:
        ref_file = fh.read()

    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    radar.fields["PH"]["_UF_scale_factor"] = 10
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]

    in_mem = StringIO()
    volume_start = netCDF4.num2date(radar.time["data"][0], radar.time["units"])
    volume_start -= datetime.timedelta(seconds=8)
    write_uf(
        in_mem,
        radar,
        uf_field_names=FIELD_MAPPING,
        field_write_order=field_write_order,
        volume_start=volume_start,
        templates_extra=TEMPLATES_EXTRA,
    )
    in_mem.seek(0)
    tst_file = in_mem.read()
    tst_file = tst_file[:208] + b"\x00\x00" + tst_file[210:]  # DZ edit_code
    tst_file = tst_file[:5700] + b"\x00\x00" + tst_file[5702:]  # ZT edit_code

    assert ref_file == tst_file


def test_complete_file_standard_names():

    with open(pyart.testing.UF_FILE, "rb") as fh:
        ref_file = fh.read()

    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    radar.fields["differential_phase"]["_UF_scale_factor"] = 10

    field_write_order = [
        "reflectivity",
        "velocity",
        "spectrum_width",
        "corrected_reflectivity",
        "total_power",
        "corrected_differential_reflectivity",
        "differential_reflectivity",
        "cross_correlation_ratio",
        "differential_phase",
        "specific_differential_phase",
        "normalized_coherent_power",
        "radar_echo_classification",
    ]
    in_mem = StringIO()
    volume_start = netCDF4.num2date(radar.time["data"][0], radar.time["units"])
    volume_start -= datetime.timedelta(seconds=8)
    write_uf(
        in_mem,
        radar,
        field_write_order=field_write_order,
        volume_start=volume_start,
        templates_extra=TEMPLATES_EXTRA,
    )
    in_mem.seek(0)
    tst_file = in_mem.read()
    tst_file = tst_file[:208] + b"\x00\x00" + tst_file[210:]  # DZ edit_code
    tst_file = tst_file[:5700] + b"\x00\x00" + tst_file[5702:]  # ZT edit_code

    assert ref_file == tst_file


def test_write_radar_field_names():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    in_mem = StringIO()
    write_uf(in_mem, radar, radar_field_names=True)
    assert in_mem.tell() == 16648


def test_write_defaults():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    in_mem = StringIO()
    write_uf(in_mem, radar)
    assert in_mem.tell() == 16648


def test_write_real_file():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    with pyart.testing.InTemporaryDirectory():
        write_uf("test.uf", radar)
        f = open("test.uf", "rb")
        buf = f.read()
        f.close()
        assert len(buf) == 16648


def test_templates_extra():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]

    templates_extra = {
        "mandatory_header": {
            "radar_name": b"xsapr-sg",
            "site_name": b"xsapr-sg",
            "generation_year": 15,
            "generation_month": 8,
            "generation_day": 19,
            "generation_facility_name": b"RSLv1.48",
        },
        "optional_header": {
            "project_name": b"TRMMGVUF",
            "tape_name": b"RADAR_UF",
        },
        "field_header": {
            "threshold_data": b"XX",
        },
    }
    ufraycreator = UFRayCreator(
        radar, FIELD_MAPPING, field_write_order, templates_extra=templates_extra
    )
    field_header = ufraycreator.field_header_template
    assert field_header["threshold_data"] == b"XX"
    man_header = ufraycreator.mandatory_header_template
    assert man_header["radar_name"] == b"xsapr-sg"
    optional_header = ufraycreator.optional_header_template
    assert optional_header["project_name"] == b"TRMMGVUF"


def test_missing_instrument_parameters():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    radar.instrument_parameters = None
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]
    ufraycreator = UFRayCreator(radar, FIELD_MAPPING, field_write_order)

    field_header = ufraycreator.field_header_template
    assert field_header["beam_width_h"] == UF_MISSING_VALUE
    assert field_header["beam_width_v"] == UF_MISSING_VALUE
    assert field_header["bandwidth"] == UF_MISSING_VALUE
    assert field_header["wavelength_cm"] == UF_MISSING_VALUE

    ufraycreator.make_field_header(999, 0, 100)
    field_header = ufraycreator.field_header_template
    assert field_header["pulse_width_m"] == UF_MISSING_VALUE
    assert field_header["prt_ms"] == UF_MISSING_VALUE
    assert field_header["polarization"] == 1

    bstring = ufraycreator.make_fsi_vel(0, 100)
    nyq = struct.unpack(">h", bstring[:2])[0]
    assert nyq == UF_MISSING_VALUE


def test_missing_scan_rate():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    radar.scan_rate = None
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]
    ufraycreator = UFRayCreator(radar, FIELD_MAPPING, field_write_order)
    ufraycreator.make_mandatory_header(0)
    man_header = ufraycreator.mandatory_header_template
    assert man_header["sweep_rate"] == UF_MISSING_VALUE


def test_unknown_scan_type():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)
    radar.scan_type = "foo"
    field_write_order = [
        "DZ",
        "VR",
        "SW",
        "CZ",
        "ZT",
        "DR",
        "ZD",
        "RH",
        "PH",
        "KD",
        "SQ",
        "HC",
    ]
    ufraycreator = UFRayCreator(radar, FIELD_MAPPING, field_write_order)
    with warnings.catch_warnings(record=True) as w:
        ufraycreator.make_mandatory_header(0)
        man_header = ufraycreator.mandatory_header_template
        assert man_header["sweep_mode"] == 1
        # check that warnings was caught
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)


def test_write_exclude():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    in_mem = StringIO()
    write_uf(in_mem, radar, exclude_fields=["reflectivity"])
    assert in_mem.tell() == 15272  # 15727 = 16648 - (667+2+19) * 2


def test_map_field_to_none():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    in_mem = StringIO()
    uf_field_names = {
        "reflectivity": None,
        "velocity": "VR",
        "spectrum_width": "SW",
        "corrected_reflectivity": "CZ",
        "total_power": "ZT",
        "corrected_differential_reflectivity": "DR",
        "differential_reflectivity": "ZD",
        "cross_correlation_ratio": "RH",
        "differential_phase": "PH",
        "specific_differential_phase": "KD",
        "normalized_coherent_power": "SQ",
        "radar_echo_classification": "HC",
    }
    write_uf(in_mem, radar, uf_field_names=uf_field_names)
    assert in_mem.tell() == 15272  # 15727 = 16648 - (667+2+19) * 2


def test_map_field_missing():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    in_mem = StringIO()
    uf_field_names = {
        "reflectivity": "DZ",
        "velocity": "VR",
        "spectrum_width": "SW",
    }
    write_uf(in_mem, radar, uf_field_names=uf_field_names)
    assert in_mem.tell() == 4264  # 4264 = 16648 - (667+2+19) * 2 * 9


def test_range_start():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    radar.range["meters_to_center_of_first_gate"] += 1500.0
    in_mem = StringIO()
    write_uf(in_mem, radar)

    in_mem.seek(0)
    ufile = UFFile(in_mem)
    field_header = ufile.rays[0].field_headers[0]
    assert field_header["range_start_km"] == 1
    assert field_header["range_start_m"] == 500



"""
===========================================================
Dealias doppler velocities using the Region Based Algorithm
===========================================================

In this example doppler velocities are dealiased using the ial condition of the dealiasing,
using the region-based dealiasing algorithm in Py-ART.
"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

radar_file = get_test_data("095636.mdv")
sonde_file = get_test_data("sgpinterpolatedsondeC1.c1.20110510.000000.cdf")

# read in the data
radar = pyart.io.read_mdv(radar_file)

# read in sonde data
dt, profile = pyart.io.read_arm_sonde_vap(sonde_file, radar=radar)

# create a gate filter which specifies gates to exclude from dealiasing
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_transition()
gatefilter.exclude_invalid("velocity")
gatefilter.exclude_invalid("reflectivity")
gatefilter.exclude_outside("reflectivity", 0, 80)

# perform dealiasing
dealias_data = pyart.correct.dealias_region_based(radar, gatefilter=gatefilter)
radar.add_field("corrected_velocity", dealias_data)

# create a plot of the first and sixth sweeps
fig = plt.figure(figsize=(15, 10))
ax1 = fig.add_subplot(221)
display = pyart.graph.RadarDisplay(radar)
display.plot(
    "velocity",
    0,
    vmin=-16,
    vmax=16,
    ax=ax1,
    colorbar_label="",
    title="Raw Doppler Velocity, First Sweep",
)

ax2 = fig.add_subplot(222)
display.plot(
    "corrected_velocity",
    0,
    vmin=-40,
    vmax=40,
    colorbar_label="",
    ax=ax2,
    title="Corrected Doppler Velocity, First Sweep",
)

ax3 = fig.add_subplot(223)
display = pyart.graph.RadarDisplay(radar)
display.plot(
    "velocity",
    5,
    vmin=-16,
    vmax=16,
    colorbar_label="",
    ax=ax3,
    title="Raw Doppler Velocity, Sixth Sweep",
)

ax4 = fig.add_subplot(224)
display.plot_ppi(
    "corrected_velocity",
    5,
    vmin=-40,
    vmax=40,
    colorbar_label="",
    ax=ax4,
    title="Corrected Doppler Velocity, Sixth Sweep",
)
plt.suptitle("Velocity dealiasing using Py-ART", fontsize=16)
plt.show()



""" Unit Tests for Py-ART's graph/radarmapdisplay.py module. """
# execute this script to create figure_plot_radar_display_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Currently this test only determines if files
# can be created, not that they are correct.

import matplotlib.pyplot as plt
import numpy as np
import pytest

import pyart


# Top level Figure generating tests
@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay._CARTOPY_AVAILABLE,
    reason="Cartopy is not installed.",
)
def test_radarmapdisplay_cartopy_ppi(outfile=None):
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar, shift=(0.1, 0.0))
    display.plot_ppi_map(
        "reflectivity_horizontal",
        0,
        colorbar_flag=True,
        title="Fancy PPI Map",
        mask_tuple=("reflectivity_horizontal", -100),
        min_lon=-100,
        max_lon=-93,
        min_lat=33,
        max_lat=38,
        mask_outside=True,
    )
    display.plot_point(-95, 35, label_text="TEXT")
    display.plot_range_rings([15, 30])
    display.plot_line_geo(np.array([-95, -95]), np.array([33, 38]))
    if outfile:
        plt.savefig(outfile)
    plt.close()


@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay._CARTOPY_AVAILABLE,
    reason="Cartopy is not installed.",
)
def test_radarmapdisplay_cartopy_preexisting_ax(outfile=None):
    import cartopy
    from cartopy.io.img_tiles import Stamen

    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar, shift=(0.1, 0.0))
    fig = plt.figure()
    ax = plt.axes(projection=cartopy.crs.PlateCarree())
    ax.add_image(Stamen("terrain-background"), 6)
    display.plot_ppi_map("reflectivity_horizontal", 0, ax=ax, embellish=False)
    if outfile:
        fig.savefig(outfile)
    plt.close()


# Tests of methods, these tests do not generate figures
#  XXX Determine why this is failing
# @pytest.mark.skipif(not pyart.graph.radarmapdisplay_cartopy._CARTOPY_AVAILABLE,
#                    reason="Cartopy is not installed.")
# def test_radarmapdisplay_cartopy_auto_range():
# test the auto_range=True function
#    import cartopy
#    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
#    display = pyart.graph.RadarMapDisplayCartopy(radar, shift=(0.1, 0.0))
#    display.plot_ppi_map('reflectivity_horizontal')
#    extent = display.ax.get_extent(cartopy.crs.PlateCarree())
#    assert round(extent[3], 2) == 36.85
#    assert round(extent[2], 2) == 36.13
#    assert round(extent[1], 2) == -97.15
#    assert round(extent[0], 2) == -98.04
#    plt.close()


@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay._CARTOPY_AVAILABLE,
    reason="Cartopy is not installed.",
)
def test_error_raising():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplay(radar, shift=(0.1, 0.0))
    # no cartopy
    pytest.raises(ValueError, display.plot_range_ring, 10)


if __name__ == "__main__":
    test_radarmapdisplay_cartopy_ppi("figure_radarmapdisplay_cartopy_ppi.png")



"""
=================================
Choose a Colormap for your Plot
=================================

This is an example of what colormaps are available in Py-ART,
and how to add them to your own plots.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np

import pyart
from pyart.testing import get_test_data

######################################
# **Plot the available colormaps**
#
# Let's see which colormaps are available directly from Py-ART!
# We use a helper function from matplotlib to plot this.

# Setup some helper functions and ranges to visualize our colormaps, from matplotlib
gradient = np.linspace(0, 1, 256)
gradient = np.vstack((gradient, gradient))


def plot_color_gradients(cmap_category, cmap_list):
    # Create figure and adjust figure height to number of colormaps
    nrows = len(cmap_list)
    figh = 0.35 + 0.15 + (nrows + (nrows - 1) * 0.1) * 0.22
    fig, axs = plt.subplots(nrows=nrows, figsize=(6.4, figh))
    fig.subplots_adjust(top=1 - 0.35 / figh, bottom=0.15 / figh, left=0.4, right=0.99)

    axs[0].set_title(cmap_category + " Colormaps", fontsize=14)

    for ax, cmap_name in zip(axs, cmap_list):
        ax.imshow(gradient, aspect="auto", cmap=f"pyart_{cmap_name}")
        ax.text(
            -0.01,
            0.5,
            f"pyart_{cmap_name}",
            va="center",
            ha="right",
            fontsize=10,
            transform=ax.transAxes,
        )

    # Turn off *all* ticks & spines, not just the ones with colormaps.
    for ax in axs:
        ax.set_axis_off()


######################################
# **Colorblind Friendly Colormaps**
#
# We recommend starting with these colorblind friendly colormaps.
# These colormaps are the most inclusive, and should be used where
# possible.

plot_color_gradients(
    "Colorblind Friendly",
    ["LangRainbow12", "HomeyerRainbow", "balance", "ChaseSpectral", "SpectralExtended"],
)

######################################
# **Perceptually Uniform Colormaps**
#
# More generally, perceptually uniform colormaps are colormaps where
# the lightness value increases monotonically through the colormaps.
plot_color_gradients(
    "Sequential",
    [
        "Bu10",
        "Bu7",
        "Gray5",
        "Gray9",
    ],
)

######################################
# **Diverging Colormaps**
#
# Diverging colormaps are helpful when showing positive and negative
# values. This is when the 0 value is meaningful (ex. velocity)
plot_color_gradients(
    "Diverging",
    [
        "BlueBrown11",
        "BrBu10",
        "BrBu12",
        "BuDOr12",
        "BuDOr18",
        "BuDRd12",
        "BuDRd18",
        "BuGr14",
        "BuGy8",
        "BuOr10",
        "BuOr12",
        "BuOr8",
        "BuOrR14",
        "GrMg16",
        "RdYlBu11b",
    ],
)

######################################
# **Field-Specific Colormaps**
#
# There are some colormaps that useful for specific fields, such as
# "BlueBrown10" for terrain, or NWSRef for the National Weather Service
# reflectivity field
plot_color_gradients(
    "Field-specific ",
    [
        "BlueBrown10",
        "Carbone11",
        "Carbone17",
        "Carbone42",
        "Cat12",
        "EWilson17",
        "NWSRef",
        "NWSVel",
        "NWS_SPW",
        "PD17",
        "RRate11",
        "RefDiff",
        "SCook18",
        "StepSeq25",
        "SymGray12",
        "Theodore16",
        "Wild25",
    ],
)


######################################
# **Plot Using a Colormap from Matplotlib**
#
# Now, we can apply one of these colorbars to plot and
# compare to a colormap from matplotlib, starting with the matplotlib example.

# Read in a sample cfradial file
radar_file = get_test_data("swx_20120520_0641.nc")
radar = pyart.io.read(radar_file)

# Setup a display to plot the data
display = pyart.graph.RadarDisplay(radar)

# Start by plotting a regular matplotlib colormap (Spectral_r)
display.plot("reflectivity_horizontal", vmin=-32, vmax=64.0, cmap="Spectral_r")

######################################
# **Plot Using a Colormap from Py-ART**
#
# Let's use one of our Py-ART's colorbars now! We need
# to remember to add the pyart_ string in front of the
# colormap, as shown below.
# Setup a display to plot the data

display = pyart.graph.RadarDisplay(radar)

# Start by plotting a regular matplotlib colormap (Spectral_r)
display.plot(
    "reflectivity_horizontal", vmin=-32, vmax=64.0, cmap="pyart_HomeyerRainbow"
)



""" Unit Tests for Py-ART's correct/dealias.py module. """

# python test_dealias.py
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

import pyart


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_dealias_sounding():
    radar, dealias_vel = perform_dealias()
    assert_allclose(
        dealias_vel["data"][13, :27],
        [
            0.5,
            1.5,
            2.5,
            3.5,
            4.5,
            5.5,
            6.5,
            7.5,
            8.5,
            9.5,
            10.5,
            11.5,
            12.5,
            13.5,
            12.5,
            11.5,
            10.5,
            9.5,
            8.5,
            7.5,
            6.5,
            5.5,
            4.5,
            3.5,
            2.5,
            1.5,
            0.5,
        ],
    )
    assert dealias_vel["data"][13, 46] is np.ma.masked
    assert dealias_vel["data"][13, 47] is np.ma.masked


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_set_limits():

    radar, dealias_vel = perform_dealias(set_limits=True)
    assert "valid_min" in dealias_vel
    assert_almost_equal(dealias_vel["valid_min"], -30.0)
    assert "valid_max" in dealias_vel
    assert_almost_equal(dealias_vel["valid_max"], 30.0)

    radar, dealias_vel = perform_dealias(set_limits=False)
    assert "valid_min" not in dealias_vel
    assert "valid_max" not in dealias_vel


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_dealias_sounding_keep_original():
    radar, dealias_vel = perform_dealias(True)
    assert_allclose(
        dealias_vel["data"][13, :27],
        [
            0.5,
            1.5,
            2.5,
            3.5,
            4.5,
            5.5,
            6.5,
            7.5,
            8.5,
            9.5,
            10.5,
            11.5,
            12.5,
            13.5,
            12.5,
            11.5,
            10.5,
            9.5,
            8.5,
            7.5,
            6.5,
            5.5,
            4.5,
            3.5,
            2.5,
            1.5,
            0.5,
        ],
    )
    assert dealias_vel["data"][13, 46] == -7.5
    assert dealias_vel["data"][13, 47] == 8.5


def perform_dealias(keep_original=False, **kwargs):
    """Perform velocity dealiasing on reference data."""
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be dealiased.
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    height = np.linspace(150, 250, 10).astype("float32")
    speed = np.ones((10), dtype="float32") * 0.5
    direction = np.ones((10), dtype="float32") * 5.0
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, sonde_profile=profile, keep_original=keep_original, **kwargs
    )
    return radar, dealias_vel


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_dealias_last_radar_and_sounding_with_profile():
    radar = pyart.testing.make_velocity_aliased_radar()
    last_radar = pyart.testing.make_velocity_aliased_radar(False)
    height = np.linspace(150, 250, 10).astype("float32")
    speed = np.ones((10), dtype="float32") * 0.5
    direction = np.ones((10), dtype="float32") * 5.0
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, sonde_profile=profile, last_radar=last_radar, last_vel_field="velocity"
    )
    assert_allclose(
        dealias_vel["data"][13, :27],
        [
            0.5,
            1.5,
            2.5,
            3.5,
            4.5,
            5.5,
            6.5,
            7.5,
            8.5,
            9.5,
            10.5,
            11.5,
            12.5,
            13.5,
            12.5,
            11.5,
            10.5,
            9.5,
            8.5,
            7.5,
            6.5,
            5.5,
            4.5,
            3.5,
            2.5,
            1.5,
            0.5,
        ],
    )
    return


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_dealias_last_radar_and_sounding():
    radar = pyart.testing.make_velocity_aliased_radar()
    last_radar = pyart.testing.make_velocity_aliased_radar(False)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, last_radar=last_radar, last_vel_field="velocity"
    )
    assert_allclose(
        dealias_vel["data"][13, :27],
        [
            0.5,
            1.5,
            2.5,
            3.5,
            4.5,
            5.5,
            6.5,
            7.5,
            8.5,
            9.5,
            10.5,
            11.5,
            12.5,
            13.5,
            12.5,
            11.5,
            10.5,
            9.5,
            8.5,
            7.5,
            6.5,
            5.5,
            4.5,
            3.5,
            2.5,
            1.5,
            0.5,
        ],
    )
    return


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_error_raising():
    # ValueError when no sounding or last_radar provided

    radar = pyart.testing.make_velocity_aliased_radar()
    pytest.raises(ValueError, pyart.correct.dealias_fourdd, radar)
    return


@pytest.mark.skipif(
    not pyart.correct.dealias._FOURDD_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_segmentation_fault_error():
    # See GitHub issue #571

    # ValueError when sounding is > 999
    radar = pyart.testing.make_velocity_aliased_radar()
    height = np.arange(2000)
    speed = np.zeros((2000,))
    direction = np.zeros((2000,))
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    pytest.raises(
        ValueError, pyart.correct.dealias_fourdd, radar, sonde_profile=profile
    )
    return


if __name__ == "__main__":

    radar, dealias_vel = perform_dealias()
    radar.fields["dealiased_velocity"] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields["velocity"]["data"][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields["dealiased_velocity"]["data"][13])

    # create plot
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi("velocity", 0, ax=ax1, colorbar_flag=False, title="", vmin=-10, vmax=10)
    rd.plot_ppi(
        "dealiased_velocity",
        0,
        ax=ax2,
        colorbar_flag=False,
        title="",
        vmin=-10,
        vmax=20,
    )
    fig.savefig("dealias_plot.png")



"""
Input/output routines common to many file formats.

"""

import bz2
import gzip

import fsspec
import netCDF4
import numpy as np


def prepare_for_read(filename, storage_options={"anon": True}):
    """
    Return a file like object read for reading.

    Open a file for reading in binary mode with transparent decompression of
    Gzip and BZip2 files. The resulting file-like object should be closed.

    Parameters
    ----------
    filename : str or file-like object
        Filename or file-like object which will be opened. File-like objects
        will not be examined for compressed data.

    storage_options : dict, optional
        Parameters passed to the backend file-system such as Google Cloud Storage,
        Amazon Web Service S3.

    Returns
    -------
    file_like : file-like object
        File like object from which data can be read.

    """
    # if a file-like object was provided, return
    if hasattr(filename, "read"):  # file-like object
        return filename

    # look for compressed data by examining the first few bytes
    fh = fsspec.open(filename, mode="rb", compression="infer", **storage_options).open()
    magic = fh.read(3)
    fh.close()

    # If the data is still compressed, use gunzip/bz2 to uncompress the data
    if magic.startswith(b"\x1f\x8b"):
        return gzip.GzipFile(filename, "rb")

    if magic.startswith(b"BZh"):
        return bz2.BZ2File(filename, "rb")

    return fsspec.open(
        filename, mode="rb", compression="infer", **storage_options
    ).open()


def stringarray_to_chararray(arr, numchars=None):
    """
    Convert an string array to a character array with one extra dimension.

    Parameters
    ----------
    arr : array
        Array with numpy dtype 'SN', where N is the number of characters
        in the string.

    numchars : int
        Number of characters used to represent the string. If numchar > N
        the results will be padded on the right with blanks. The default,
        None will use N.

    Returns
    -------
    chararr : array
        Array with dtype 'S1' and shape = arr.shape + (numchars, ).

    """
    carr = netCDF4.stringtochar(arr)
    if numchars is None:
        return carr

    arr_numchars = carr.shape[-1]
    if numchars <= arr_numchars:
        raise ValueError("numchars must be >= %i" % (arr_numchars))
    chararr = np.zeros(arr.shape + (numchars,), dtype="S1")
    chararr[..., :arr_numchars] = carr[:]
    return chararr


def _test_arguments(dic):
    """Issue a warning if receive non-empty argument dict."""
    if dic:
        import warnings

        warnings.warn("Unexpected arguments: %s" % dic.keys())


def make_time_unit_str(dtobj):
    """Return a time unit string from a datetime object."""
    return "seconds since " + dtobj.strftime("%Y-%m-%dT%H:%M:%SZ")



""" Unit Tests for Py-ART's correct/attenuation.py module. """

# python test_phase_attenuation.py
# to recreate the reference_rays.npz file.

import os

import numpy as np
from numpy.testing import assert_allclose

import pyart

PATH = os.path.dirname(__file__)
REFERENCE_RAYS_FILE = os.path.join(PATH, "attenuation_rays.npz")
REFERENCE_RAYS_FILE_ZPHI = os.path.join(PATH, "attenuation_rays_zphi.npz")
REFERENCE_RAYS_FILE_PHILINEAR = os.path.join(PATH, "attenuation_rays_philinear.npz")


def test_attenuation():
    spec_at, cor_z = perform_attenuation()
    ref = np.load(REFERENCE_RAYS_FILE)
    assert_allclose(ref["spec_at"], spec_at["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["cor_z"], cor_z["data"].data, rtol=1e-2, atol=1e-3)


def test_attenuation_gatefilter():
    spec_at, cor_z = perform_attenuation(use_gatefilter=True)
    ref = np.load(REFERENCE_RAYS_FILE)
    assert_allclose(ref["spec_at"], spec_at["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["cor_z"], cor_z["data"].data, rtol=1e-2, atol=1e-3)


def perform_attenuation_zphi():
    """Perform attenuation correction on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    a = radar.fields["reflectivity"]["data"]
    radar.fields["reflectivity"]["data"] = np.ma.array(a)
    (
        spec_at,
        pia_dict,
        cor_z,
        spec_diff_at,
        pida_dict,
        cor_zdr,
    ) = pyart.correct.calculate_attenuation_zphi(
        radar,
        a_coef=0.06,
        beta=0.8,
        fzl=4000.0,
        c=0.15917,
        d=1.0804,
        doc=0.0,
        temp_ref="fixed_fzl",
    )
    return spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr


def perform_attenuation_philinear():
    """Perform attenuation correction on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    a = radar.fields["reflectivity"]["data"]
    radar.fields["reflectivity"]["data"] = np.ma.array(a)
    (
        spec_at,
        pia_dict,
        cor_z,
        spec_diff_at,
        pida_dict,
        cor_zdr,
    ) = pyart.correct.calculate_attenuation_philinear(
        radar, pia_coef=0.06, pida_coef=0.8, fzl=4000.0, doc=0.0, temp_ref="fixed_fzl"
    )
    return spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr


def perform_attenuation(use_gatefilter=False):
    """Perform attenuation correction on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    if use_gatefilter:
        gatefilter = pyart.filters.GateFilter(radar)
    else:
        gatefilter = None
    a = radar.fields["reflectivity"]["data"]
    radar.fields["reflectivity"]["data"] = np.ma.array(a)
    spec_at, cor_z = pyart.correct.calculate_attenuation(
        radar, 0.0, gatefilter=gatefilter
    )
    return spec_at, cor_z


def save_reference_rays(spec_at, cor_z):
    """Save the phase processed rays to REFERENCE_RAY_FILE."""
    np.savez(REFERENCE_RAYS_FILE, spec_at=spec_at["data"], cor_z=cor_z["data"].data)


def save_reference_rays_zphi(
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr
):
    """Save the phase processed rays to REFERENCE_RAY_FILE_ZPHI."""
    np.savez(
        REFERENCE_RAYS_FILE_ZPHI,
        spec_at=spec_at["data"],
        cor_z=cor_z["data"].data,
        pia_dict=pia_dict["data"],
        spec_diff_at=spec_diff_at["data"],
        pida_dict=pida_dict["data"],
        cor_zdr=cor_zdr["data"],
    )


def save_reference_rays_philinear(
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr
):
    """Save the phase processed rays to REFERENCE_RAY_FILE_PHILINEAR."""
    np.savez(
        REFERENCE_RAYS_FILE_PHILINEAR,
        spec_at=spec_at["data"],
        cor_z=cor_z["data"].data,
        pia_dict=pia_dict["data"],
        spec_diff_at=spec_diff_at["data"],
        pida_dict=pida_dict["data"],
        cor_zdr=cor_zdr["data"],
    )


if __name__ == "__main__":
    spec_at, cor_z = perform_attenuation()
    save_reference_rays(spec_at, cor_z)
    (
        spec_at,
        pia_dict,
        cor_z,
        spec_diff_at,
        pida_dict,
        cor_zdr,
    ) = perform_attenuation_zphi()
    save_reference_rays_zphi(spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr)
    (
        spec_at,
        pia_dict,
        cor_z,
        spec_diff_at,
        pida_dict,
        cor_zdr,
    ) = perform_attenuation_philinear()
    save_reference_rays_philinear(
        spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr
    )
    print("Reference rays saved")


def test_specific_diff_attenuation():
    """Perform attenuation correction on a single ray radar."""
    (
        spec_at,
        pia_dict,
        cor_z,
        spec_diff_at,
        pida_dict,
        cor_zdr,
    ) = perform_attenuation_zphi()
    ref = np.load(REFERENCE_RAYS_FILE_ZPHI)
    assert_allclose(ref["spec_at"], spec_at["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["cor_z"], cor_z["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["pia_dict"], pia_dict["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["spec_diff_at"], spec_diff_at["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["pida_dict"], pida_dict["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["cor_zdr"], cor_zdr["data"], rtol=1e-2, atol=1e-3)

    (
        spec_at,
        pia_dict,
        cor_z,
        spec_diff_at,
        pida_dict,
        cor_zdr,
    ) = perform_attenuation_philinear()
    ref = np.load(REFERENCE_RAYS_FILE_PHILINEAR)
    assert_allclose(ref["spec_at"], spec_at["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["cor_z"], cor_z["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["pia_dict"], pia_dict["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["spec_diff_at"], spec_diff_at["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["pida_dict"], pida_dict["data"], rtol=1e-2, atol=1e-3)
    assert_allclose(ref["cor_zdr"], cor_zdr["data"], rtol=1e-2, atol=1e-3)



""" Unit Tests for Py-ART's io/mdv.py module. """

import bz2
from io import BytesIO

import pytest

import pyart


def test_autoread_mdv():
    radar = pyart.io.read(pyart.testing.MDV_PPI_FILE)
    source = radar.metadata["source"]
    assert source == "MDV radar volume file created by Dsr2Vol."

    radar = pyart.io.read(pyart.testing.MDV_RHI_FILE)
    source = radar.metadata["source"]
    assert source == "MDV radar volume file created by Dsr2Vol."


def test_autoread_sigmet():
    radar = pyart.io.read(pyart.testing.SIGMET_PPI_FILE, use_rsl=False)
    assert radar.metadata["original_container"] == "sigmet"

    radar = pyart.io.read(pyart.testing.SIGMET_RHI_FILE, use_rsl=False)
    assert radar.metadata["original_container"] == "sigmet"


@pytest.mark.skipif(
    not pyart.io.rsl._RSL_AVAILABLE, reason="TRMM RSL is not installed."
)
def test_autoread_sigmet_rsl():
    radar = pyart.io.read(pyart.testing.SIGMET_PPI_FILE, use_rsl=True)
    assert radar.metadata["original_container"] == "rsl"

    radar = pyart.io.read(pyart.testing.SIGMET_RHI_FILE, use_rsl=True)
    assert radar.metadata["original_container"] == "rsl"


def test_autoread_cfradial():
    radar = pyart.io.read(pyart.testing.CFRADIAL_PPI_FILE)
    assert radar.metadata["comment"] == "none"

    radar = pyart.io.read(pyart.testing.CFRADIAL_RHI_FILE)
    assert radar.metadata["comment"] == "none"


def test_autoread_nexrad_archive():
    radar = pyart.io.read(
        pyart.testing.NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE, delay_field_loading=True
    )
    assert radar.metadata["original_container"] == "NEXRAD Level II"

    radar = pyart.io.read(
        pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE, delay_field_loading=True
    )
    assert radar.metadata["original_container"] == "NEXRAD Level II"


def test_autoread_nexrad_cdm():
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_nexrad.nc"
        with open(tmpfile, "wb") as f:
            f.write(bz2.BZ2File(pyart.testing.NEXRAD_CDM_FILE).read())
        radar = pyart.io.read(tmpfile)
        assert radar.metadata["original_container"] == "NEXRAD Level II"


def test_autoread_nexrad_level3():
    radar = pyart.io.read(pyart.testing.NEXRAD_LEVEL3_MSG19)
    assert radar.metadata["original_container"] == "NEXRAD Level 3"


def test_autoread_raises():
    f = BytesIO(b"0000000000000000000")
    pytest.raises(TypeError, pyart.io.read, f)


headers = [
    (b"\x00\x00\x03\xf8\x00\x007>\x00\x00\x00\x01", "MDV"),
    (b"\x89HDF\r\n\x1a\n\x02\x08\x08\x00", "NETCDF4"),
    (b"CDF", "NETCDF3"),
    (b"AR2V0006.501", "WSR88D"),
    (b"SDUS54 KBMX ", "NEXRADL3"),
    (b"\x1b\x00\x08\x00\x00\x08\xb7\x07\x00\x00\x00\x00", "SIGMET"),
    (b"BZh91AY&SY\xbd\x12", "BZ2"),
    (b"UF", "UF"),  # not from a real file
    (b"SSWB", "DORADE"),  # not from a real file
    (b"RSL", "RSL"),  # not from a real file
    (b"\x0e\x03\x13\x01", "HDF4"),  # not from a real file
    (b"000000000000", "UNKNOWN"),
]


@pytest.mark.parametrize("i", headers)
def test_determine_filetype(i):
    string, filetype = i
    check_filetype.description = "determine filetype: %s" % (filetype)
    check_filetype(string, filetype)


def check_filetype(string, filetype):
    f = BytesIO(string)
    assert pyart.io.auto_read.determine_filetype(f) == filetype



"""
Reading of Universal format (UF) files.

"""

import warnings

import numpy as np
from netCDF4 import date2num

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import _test_arguments, make_time_unit_str, prepare_for_read
from .uffile import UFFile

_LIGHT_SPEED = 2.99792458e8  # speed of light in meters per second
_UF_SWEEP_MODES = {
    0: "calibration",
    1: "ppi",
    2: "coplane",
    3: "rhi",
    4: "vpt",
    5: "target",
    6: "manual",
    7: "idle",
    8: "ppi",  # RadX used this to indicate surveillance PPI scans
}

_SWEEP_MODE_STR = {
    "calibration": "calibration",
    "ppi": "azimuth_surveillance",
    "coplane": "coplane",
    "rhi": "rhi",
    "vpt": "vertical_pointing",
    "target": "pointing",
    "manual": "manual",
    "idle": "idle",
}


def read_uf(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    delay_field_loading=False,
    **kwargs
):
    """
    Read a UF File.

    Parameters
    ----------
    filename : str or file-like
        Name of Universal format file to read data from.
    field_names : dict, optional
        Dictionary mapping UF data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduce any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        This option is not implemented in the function but included for
        compatibility.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata(
        "uf",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # Open UF file and get handle
    ufile = UFFile(prepare_for_read(filename))
    first_ray = ufile.rays[0]

    # time
    dts = ufile.get_datetimes()
    units = make_time_unit_str(min(dts))
    time = filemetadata("time")
    time["units"] = units
    time["data"] = date2num(dts, units).astype("float32")

    # range
    _range = filemetadata("range")
    # assume that the number of gates and spacing from the first ray is
    # representative of the entire volume
    field_header = first_ray.field_headers[0]
    ngates = field_header["nbins"]
    step = field_header["range_spacing_m"]
    # this gives distances to the center of each gate, remove step/2 for start
    start = (
        field_header["range_start_km"] * 1000.0
        + field_header["range_start_m"]
        + step / 2.0
    )
    _range["data"] = np.arange(ngates, dtype="float32") * step + start
    _range["meters_to_center_of_first_gate"] = start
    _range["meters_between_gates"] = step

    # latitude, longitude and altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")
    lat, lon, height = first_ray.get_location()
    latitude["data"] = np.array([lat], dtype="float64")
    longitude["data"] = np.array([lon], dtype="float64")
    altitude["data"] = np.array([height], dtype="float64")

    # metadata
    metadata = filemetadata("metadata")
    metadata["original_container"] = "UF"
    metadata["site_name"] = first_ray.mandatory_header["site_name"]
    metadata["radar_name"] = first_ray.mandatory_header["radar_name"]

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")
    sweep_start_ray_index["data"] = ufile.first_ray_in_sweep
    sweep_end_ray_index["data"] = ufile.last_ray_in_sweep

    # sweep number
    sweep_number = filemetadata("sweep_number")
    sweep_number["data"] = np.arange(ufile.nsweeps, dtype="int32")

    # scan_type
    scan_type = _get_scan_type(first_ray)

    # sweep_mode
    sweep_mode = filemetadata("sweep_mode")
    sweep_mode["data"] = np.array(
        ufile.nsweeps * [_SWEEP_MODE_STR[scan_type]], dtype="S"
    )

    # elevation
    elevation = filemetadata("elevation")
    elevation["data"] = ufile.get_elevations()

    # azimuth
    azimuth = filemetadata("azimuth")
    azimuth["data"] = ufile.get_azimuths()

    # fixed_angle
    fixed_angle = filemetadata("fixed_angle")
    fixed_angle["data"] = ufile.get_sweep_fixed_angles()

    # fields
    fields = {}
    for uf_field_number, uf_field_dic in enumerate(first_ray.field_positions):
        uf_field_name = uf_field_dic["data_type"].decode("ascii")
        field_name = filemetadata.get_field_name(uf_field_name)
        if field_name is None:
            continue
        field_dic = filemetadata(field_name)
        field_dic["data"] = ufile.get_field_data(uf_field_number)
        field_dic["_FillValue"] = get_fillvalue()
        fields[field_name] = field_dic

    # instrument_parameters
    instrument_parameters = _get_instrument_parameters(ufile, filemetadata)

    # scan rate
    scan_rate = filemetadata("scan_rate")
    scan_rate["data"] = ufile.get_sweep_rates()

    ufile.close()
    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        scan_rate=scan_rate,
        instrument_parameters=instrument_parameters,
    )


def _get_scan_type(ufray):
    """Ruturn the scan type of a UF ray."""
    uf_sweep_mode = ufray.mandatory_header["sweep_mode"]
    if uf_sweep_mode in _UF_SWEEP_MODES:
        return _UF_SWEEP_MODES[uf_sweep_mode]
    else:
        warnings.warn("Unknown sweep mode, defaulting to ppi")
        return "ppi"


def _get_instrument_parameters(ufile, filemetadata):
    """Return a dictionary containing instrument parameters."""

    # pulse width
    pulse_width = filemetadata("pulse_width")
    pulse_width["data"] = ufile.get_pulse_widths() / _LIGHT_SPEED  # m->sec

    # assume that the parameters in the first ray represent the beam widths,
    # bandwidth and frequency in the entire volume
    first_ray = ufile.rays[0]
    field_header = first_ray.field_headers[0]
    beam_width_h = field_header["beam_width_h"] / 64.0
    beam_width_v = field_header["beam_width_v"] / 64.0
    bandwidth = field_header["bandwidth"] / 16.0 * 1.0e6
    wavelength_cm = field_header["wavelength_cm"] / 64.0
    if wavelength_cm == 0:
        warnings.warn("Invalid wavelength, frequency set to default value.")
        wavelength_hz = 9999.0
    else:
        wavelength_hz = _LIGHT_SPEED / (wavelength_cm / 100.0)

    # radar_beam_width_h
    radar_beam_width_h = filemetadata("radar_beam_width_h")
    radar_beam_width_h["data"] = np.array([beam_width_h], dtype="float32")

    # radar_beam_width_v
    radar_beam_width_v = filemetadata("radar_beam_width_v")
    radar_beam_width_v["data"] = np.array([beam_width_v], dtype="float32")

    # radar_receiver_bandwidth
    radar_receiver_bandwidth = filemetadata("radar_receiver_bandwidth")
    radar_receiver_bandwidth["data"] = np.array([bandwidth], dtype="float32")

    # polarization_mode
    polarization_mode = filemetadata("polarization_mode")
    polarization_mode["data"] = ufile.get_sweep_polarizations()

    # frequency
    frequency = filemetadata("frequency")
    frequency["data"] = np.array([wavelength_hz], dtype="float32")

    # prt
    prt = filemetadata("prt")
    prt["data"] = ufile.get_prts() / 1e6  # us->sec

    instrument_parameters = {
        "pulse_width": pulse_width,
        "radar_beam_width_h": radar_beam_width_h,
        "radar_beam_width_v": radar_beam_width_v,
        "radar_receiver_bandwidth": radar_receiver_bandwidth,
        "polarization_mode": polarization_mode,
        "frequency": frequency,
        "prt": prt,
    }

    # nyquist velocity if defined
    nyquist_velocity = filemetadata("nyquist_velocity")
    nyquist_velocity["data"] = ufile.get_nyquists()
    if nyquist_velocity["data"] is not None:
        instrument_parameters["nyquist_velocity"] = nyquist_velocity

    return instrument_parameters



""" Unit Tests for Py-ART's graph/gridmapdisplay_basemap.py module. """
# execute this script to create figure_gridmapdisplay_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
import pytest

import pyart

RESOLUTION = "c"  # crude resolution to speed up tests


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_gridmapdisplay_simple_basemap(outfile=None):
    # test basic GridMapDisplay functionality
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_basemap(ax=ax, resolution=RESOLUTION)
    display.plot_grid("reflectivity", vmin=-5.0, vmax=35)
    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_gridmapdisplay_fancy_basemap(outfile=None):
    # test a bunch of GridMapDisplay functionality
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    display.debug = True
    fig = plt.figure()

    ax1 = fig.add_subplot(331)
    display.plot_basemap(ax=ax1, resolution=RESOLUTION)
    display.plot_grid(
        "reflectivity",
        vmin=-5.0,
        vmax=35,
        ax=ax1,
        mask_outside=True,
        axislabels_flag=True,
        axislabels=("foo", "bar"),
        title="Special title",
    )
    display.plot_crosshairs(ax=ax1, line_style="b--")

    ax2 = fig.add_subplot(332)
    display.plot_grid("reflectivity", ax=ax2, axislabels_flag=True)

    ax3 = fig.add_subplot(333)
    display.plot_colorbar(ax=ax3)

    ax4 = fig.add_subplot(334)
    display.plot_latitude_slice("reflectivity", ax=ax4, mask_outside=True)

    ax5 = fig.add_subplot(335)
    display.plot_latitude_slice("reflectivity", ax=ax5, title="Lat title")

    ax6 = fig.add_subplot(336)
    grid.fields["reflectivity"]["valid_min"] = 0.0
    grid.fields["reflectivity"]["valid_max"] = 30.0
    display.plot_longitude_slice("reflectivity", ax=ax6, mask_outside=True)

    ax7 = fig.add_subplot(337)
    display.plot_longitude_slice("reflectivity", ax=ax7, title="Lon title")

    ax8 = fig.add_subplot(338)
    del display.grid.fields["reflectivity"]["long_name"]
    display.plot_colorbar(ax=ax8)

    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_plot_basemap_not_using_corners(outfile=None):
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_basemap(ax=ax, resolution=RESOLUTION, max_lon=None, auto_range=False)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_generate_filename_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    filename = display.generate_filename("reflectivity", 0)
    assert isinstance(filename, str)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_generate_titles_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)

    title = display.generate_longitudinal_level_title("reflectivity", 0)
    assert isinstance(title, str)

    title = display.generate_latitudinal_level_title("reflectivity", 0)
    assert isinstance(title, str)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_get_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)
    display.get_basemap()


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_error_raising_basemap():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplayBasemap(grid)

    # no mappable
    pytest.raises(ValueError, display.plot_colorbar)

    # no field
    display.mappables.append(None)  # mock the mappable
    pytest.raises(ValueError, display.plot_colorbar)


if __name__ == "__main__":
    test_gridmapdisplay_simple_basemap("figure_gridmapdisplay_simple_basemap.png")
    test_gridmapdisplay_fancy_basemap("figure_gridmapdisplay_fancy_basemap.png")



""" Unit Tests for Py-ART's correct/proc_phase.py module. """

# Can also be run as a script to create a ray_plot.png file
# python test_phase_proc.py
# This file should look like the reference_ray_plot.png file

# Use:
# python test_phase_proc.py -r
# to recreate the reference_rays.npz and reference_ray_plot.png files

import os
import warnings

import numpy as np
import pytest

try:
    import cvxopt  # noqa

    cvxopt_available = True
except ImportError:
    cvxopt_available = False

try:
    import glpk  # noqa

    glpk_available = True
except ImportError:
    glpk_available = False


try:
    import cylp.cy  # noqa

    cylp_available = True
except ImportError:
    cylp_available = False

import pyart

PATH = os.path.dirname(__file__)
REFERENCE_RAYS_FILE = os.path.join(PATH, "reference_rays.npz")


@pytest.mark.skipif(not glpk_available, reason="GLPK is not installed.")
def test_phase_proc_lp_glpk():
    radar, phidp, kdp = perform_phase_processing()
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


@pytest.mark.skipif(not cvxopt_available, reason="CVXOPT is not installed.")
def test_phase_proc_lp_cvxopt():
    from cvxopt import solvers

    solvers.options["LPX_K_MSGLEV"] = 0  # supress screen output
    radar, phidp, kdp = perform_phase_processing("cvxopt")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


@pytest.mark.skipif(not cylp_available, reason="CyLP is not installed.")
def test_phase_proc_lp_cylp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing("cylp")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


@pytest.mark.skipif(not cylp_available, reason="CyLP is not installed.")
def test_phase_proc_lp_gf_cylp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing_gf("cylp")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.05
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.05
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.05
    )


@pytest.mark.skipif(not cylp_available, reason="CyLP is not installed.")
def test_phase_proc_lp_cylp_mp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing("cylp_mp")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


def _ratio(a1, a2):
    """Ratio the sum of the abs difference vs sum abs of two vectors."""
    abs_residues = np.abs(a1 - a2).sum()
    avg_abs_sum = 0.5 * np.abs(a1).sum() + 0.5 * np.abs(a2).sum()
    return abs_residues / avg_abs_sum


def perform_phase_processing(LP_solver="pyglpk"):
    """Perform LP phase processing on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    phidp, kdp = pyart.correct.phase_proc_lp(radar, 0.0, LP_solver=LP_solver)
    return radar, phidp, kdp


def perform_phase_processing_gf(LP_solver="pyglpk"):
    """Perform LP phase processing on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    my_gatefilter = pyart.filters.GateFilter(radar)
    my_gatefilter.exclude_below("normalized_coherent_power", 0.5)
    my_gatefilter.exclude_below("cross_correlation_ratio", 0.8)
    phidp, kdp = pyart.correct.phase_proc_lp_gf(
        radar,
        gatefilter=my_gatefilter,
        LP_solver=LP_solver,
        doc=15,
        ncpts=20,
        system_phase=-140.1,
    )
    return radar, phidp, kdp


def save_reference_rays(radar, phidp, kdp):
    """Save the phase processed rays to REFERENCE_RAY_FILE."""
    unfolded = radar.fields["unfolded_differential_phase"]
    np.savez(
        REFERENCE_RAYS_FILE,
        reference_phidp=phidp["data"],
        reference_kdp=kdp["data"],
        reference_unfolded_phidp=unfolded["data"],
    )


def make_plot(range_km, unfolded_phidp, refl, phidp, kdp, filename):
    """
    Make and save a plot of PhiDP, reflectivity and KDP.

    Parameters
    ----------
    range_km : array
        Gate range in kilometers.
    unfolded_phidp, refl, phidp, kdp : dict
        Dictionary containing the unfolded phiDP, reflectivity,
        LP filtered phiDP and KDP in the data key
    filename :
        Filename to save the file to.

    """

    from matplotlib import pyplot as plt

    fig = plt.figure(figsize=[10, 5])
    ax = fig.add_subplot(111)

    # filtered phidp and unfolded phidp
    (p1,) = ax.plot(range_km, phidp["data"][0], "b-")
    (p2,) = ax.plot(range_km, unfolded_phidp["data"][0], "g-")

    # set labels
    ax.set_ylim(0, 250)
    ax.set_ylabel("Differential phase shift (degrees)")
    ax.set_xlabel("Range (km)")

    # plot KDP and reflectivity on second axis
    ax2 = ax.twinx()
    (p3,) = ax2.plot(range_km, kdp["data"][0], "r-")
    (p4,) = ax2.plot(range_km, refl["data"][0] / 10.0)

    # decorate and save
    ax2.yaxis.grid(color="gray", linestyle="dashed")
    ax.legend(
        [p1, p2, p3, p4],
        ["Filtered phiDP", "Unfolded phiDP", "KDP", "Z/10.0"],
        loc="upper left",
    )
    fig.savefig(filename)


if __name__ == "__main__":
    import sys

    radar, phidp, kdp = perform_phase_processing()
    filename = "ray_plot.png"
    if sys.argv[-1] == "-r":  # regenerate reference files
        save_reference_rays(radar, phidp, kdp)
        filename = "reference_ray_plot.png"

    make_plot(
        radar.range["data"] / 1000.0,
        radar.fields["unfolded_differential_phase"],
        radar.fields["reflectivity"],
        phidp,
        kdp,
        filename,
    )



"""
Correct radar fields.

"""

# for backwards compatibility GateFilter available in the correct namespace
from ..filters.gatefilter import GateFilter, moment_based_gate_filter  # noqa
from .attenuation import calculate_attenuation  # noqa
from .attenuation import calculate_attenuation_philinear  # noqa
from .attenuation import calculate_attenuation_zphi  # noqa
from .bias_and_noise import correct_bias, correct_noise_rhohv  # noqa
from .dealias import dealias_fourdd  # noqa
from .despeckle import despeckle_field, find_objects  # noqa
from .phase_proc import phase_proc_lp, phase_proc_lp_gf  # noqa
from .region_dealias import dealias_region_based  # noqa
from .unwrap import dealias_unwrap_phase  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



""" Unit Tests for Py-ART's graph/radarmapdisplay.py module. """
# execute this script to create figure_plot_radar_display_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Current this test only determines if files can
# be created, not that they are correct.

import matplotlib.pyplot as plt
import numpy as np
import pytest

import pyart


# Top level Figure generating tests
@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_radarmapdisplay_ppi_basemap(outfile=None):
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplayBasemap(radar, shift=(0.1, 0.0))
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi_map(
        "reflectivity_horizontal",
        0,
        ax=ax,
        colorbar_flag=True,
        title="Fancy PPI Map",
        mask_tuple=("reflectivity_horizontal", -100),
        resolution="c",
        min_lon=-100,
        max_lon=-93,
        min_lat=33,
        max_lat=38,
        mask_outside=True,
    )
    display.plot_point(-95, 35, label_text="TEXT")
    display.plot_range_rings([15, 30])
    display.plot_line_geo(np.array([-95, -95]), np.array([33, 38]))
    if outfile:
        fig.savefig(outfile)
    plt.close()


# Tests of methods, these tests do not generate figures
@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_radarmapdisplay_auto_range_basemap():
    # test the auto_range=True function
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplayBasemap(radar, shift=(0.1, 0.0))
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_ppi_map("reflectivity_horizontal", ax=ax, resolution="c")
    assert round(display.basemap.latmax, 2) == 36.85
    assert round(display.basemap.latmin, 2) == 36.13
    assert round(display.basemap.lonmax, 2) == -97.15
    assert round(display.basemap.lonmin, 2) == -98.04
    plt.close()


@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_error_raising_basemap():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplayBasemap(radar, shift=(0.1, 0.0))
    # no basemap
    pytest.raises(ValueError, display.plot_range_ring, 10)


@pytest.mark.skipif(
    not pyart.graph.radarmapdisplay_basemap._BASEMAP_AVAILABLE,
    reason="Basemap is not installed.",
)
def test_radardisplay_cylindrical_proj_error_basemap():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
    display = pyart.graph.RadarMapDisplayBasemap(radar)
    pytest.raises(
        ValueError,
        display.plot_ppi_map,
        "reflectivity_horizontal",
        projection="cyl",
        resolution="c",
    )


if __name__ == "__main__":
    test_radarmapdisplay_ppi_basemap("figure_radarmapdisplay_ppi.png")



"""
Python wrapper around the RSL library.

"""

import datetime
import warnings

import numpy as np

from ..config import FileMetadata, get_fillvalue

try:
    from . import _rsl_interface

    _RSL_AVAILABLE = True
except ImportError:
    _RSL_AVAILABLE = False
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency
from ..lazydict import LazyLoadDict
from .common import make_time_unit_str


def read_rsl(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    delay_field_loading=False,
    include_fields=None,
    radar_format=None,
    callid=None,
    skip_range_check=False,
):
    """
    Read a file supported by RSL.

    Parameters
    ----------
    filename : str or RSL_radar
        Name of file whose format is supported by RSL.
    field_names : dict, optional
        Dictionary mapping RSL data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the RSL data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed. In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects.
    radar_format : str or None
        Format of the radar file. Must be 'wsr88d' or None.
    callid : str or None
        Four letter NEXRAD radar Call ID, only used when radar_format is
        'wsr88d'.
    skip_range_check : bool, optional
        True to skip check for uniform range bin location, the reported range
        locations will only be verified true for the first ray. False will
        perform the check and raise a IOError when the locations of the gates
        change between rays.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # check that TRMM RSL is available
    if not _RSL_AVAILABLE:
        raise MissingOptionalDependency(
            "Py-ART must be build with support for TRMM RSL to use the "
            "read_rsl function."
        )

    # create metadata retrieval object
    filemetadata = FileMetadata(
        "rsl",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # read the file, determine common parameters
    fillvalue = get_fillvalue()
    rslfile = _rsl_interface.RslFile(filename.encode("ascii"), radar_format, callid)
    available_vols = rslfile.available_moments()
    first_volume = rslfile.get_volume(available_vols[0])
    first_sweep = first_volume.get_sweep(0)
    first_ray = first_sweep.get_ray(0)
    nsweeps = first_volume.nsweeps

    # scan_type, sweep_mode, fixed_angle
    fixed_angle = filemetadata("fixed_angle")
    fdata = first_volume.get_sweep_fix_angles()
    fdata[fdata < 0.0] = 360.0 + fdata[fdata < 0.0]  # all angles positive
    fixed_angle["data"] = fdata
    sweep_mode = filemetadata("sweep_mode")

    if rslfile.scan_mode == 0:  # 0 = PPI, 1 = RHI
        scan_type = "ppi"
        sweep_mode["data"] = np.array(nsweeps * ["azimuth_surveillance"], dtype="S")
    else:
        scan_type = "rhi"
        sweep_mode["data"] = np.array(nsweeps * ["rhi"], dtype="S")

    # time
    time = filemetadata("time")
    t_start = datetime.datetime(
        rslfile.year,
        rslfile.month,
        rslfile.day,
        rslfile.hour,
        rslfile.minute,
        int(rslfile.sec),
        int(1e6 * (rslfile.sec % 1)),
    )
    datetimes = []
    for i in range(nsweeps):
        sweep = first_volume.get_sweep(i)
        for j in range(sweep.nrays):
            datetimes.append(sweep.get_ray(j).get_datetime())
    t_delta = [t - t_start for t in datetimes]
    sec_since_start = [
        t.seconds + t.days * 3600 * 24 + t.microseconds / 1.0e6 for t in t_delta
    ]
    time["data"] = np.array(sec_since_start, dtype=np.float64)
    time["units"] = make_time_unit_str(t_start)

    # range
    if (not skip_range_check) and (not first_volume.is_range_bins_uniform()):
        message = (
            "Range bin locations change between rays. File cannot be read "
            "with with correct range locations for all rays. "
            "To read in data reporting the ranges from the first ray set the "
            "**skip_range_check** parameter to True."
        )
        raise OSError(message)

    _range = filemetadata("range")
    gate0 = first_ray.range_bin1
    gate_size = first_ray.gate_size
    ngates = first_ray.nbins
    _range["data"] = gate0 + gate_size * np.arange(ngates, dtype="float32")
    _range["meters_to_center_of_first_gate"] = _range["data"][0]
    _range["meters_between_gates"] = np.array(gate_size, dtype="float32")

    # fields
    # transfer only those which are available
    fields = {}
    for volume_num in available_vols:
        # if a volume number is not recognized it is skipped.
        if volume_num not in VOLUMENUM2RSLNAME:
            warnings.warn("Unknown Volume Number %d" % volume_num)
            continue
        rsl_field_name = VOLUMENUM2RSLNAME[volume_num]
        field_name = filemetadata.get_field_name(rsl_field_name)
        if field_name is None:
            continue

        # create the field dictionary
        field_dic = filemetadata(field_name)
        field_dic["_FillValue"] = fillvalue
        data_extractor = _RslVolumeDataExtractor(rslfile, volume_num, fillvalue)
        if delay_field_loading:
            field_dic = LazyLoadDict(field_dic)
            field_dic.set_lazy("data", data_extractor)
        else:
            field_dic["data"] = data_extractor()
        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata("metadata")
    metadata["original_container"] = "rsl"
    rsl_dict = rslfile.get_radar_header()
    need_from_rsl_header = {
        "name": "instrument_name",
        "project": "project",
        "state": "state",
        "country": "country",
    }  # rsl_name : radar_metadata_name
    for rsl_key, metadata_key in need_from_rsl_header.items():
        metadata[metadata_key] = rsl_dict[rsl_key]

    # latitude
    latitude = filemetadata("latitude")
    lat = _dms_to_d((rsl_dict["latd"], rsl_dict["latm"], rsl_dict["lats"]))
    latitude["data"] = np.array([lat], dtype="float64")

    # longitude
    longitude = filemetadata("longitude")
    lon = _dms_to_d((rsl_dict["lond"], rsl_dict["lonm"], rsl_dict["lons"]))
    longitude["data"] = np.array([lon], dtype="float64")

    # altitude
    altitude = filemetadata("altitude")
    altitude["data"] = np.array([rsl_dict["height"]], dtype="float64")

    # sweep_number, sweep_mode, sweep_start_ray_index, sweep_end_ray_index
    sweep_number = filemetadata("sweep_number")
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")

    sweep_number["data"] = np.arange(nsweeps, dtype="int32")
    ray_count = first_volume.get_nray_array()  # array of rays in each sweep
    ssri = np.cumsum(np.append([0], ray_count[:-1])).astype("int32")
    sweep_start_ray_index["data"] = ssri
    sweep_end_ray_index["data"] = np.cumsum(ray_count).astype("int32") - 1

    # azimuth, elevation
    azimuth = filemetadata("azimuth")
    elevation = filemetadata("elevation")
    _azimuth, _elevation = first_volume.get_azimuth_and_elev_array()
    azimuth["data"] = _azimuth
    elevation["data"] = _elevation

    # instrument_parameters
    prt = filemetadata("prt")
    prt_mode = filemetadata("prt_mode")
    nyquist_velocity = filemetadata("nyquist_velocity")
    unambiguous_range = filemetadata("unambiguous_range")
    beam_width_h = filemetadata("radar_beam_width_h")
    beam_width_v = filemetadata("radar_beam_width_v")

    pm_data, nv_data, pr_data, ur_data = first_volume.get_instr_params()
    prt["data"] = pr_data
    prt_mode["data"] = pm_data.astype("S")
    nyquist_velocity["data"] = nv_data
    unambiguous_range["data"] = ur_data
    beam_width_h["data"] = np.array([first_sweep.horz_half_bw * 2.0], dtype="float32")
    beam_width_v["data"] = np.array([first_sweep.vert_half_bw * 2.0], dtype="float32")

    instrument_parameters = {
        "unambiguous_range": unambiguous_range,
        "prt_mode": prt_mode,
        "prt": prt,
        "nyquist_velocity": nyquist_velocity,
        "radar_beam_width_h": beam_width_h,
        "radar_beam_width_v": beam_width_v,
    }

    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )


def _dms_to_d(dms):
    """Degrees, minutes, seconds to degrees"""
    return dms[0] + (dms[1] + dms[2] / 60.0) / 60.0


class _RslVolumeDataExtractor:
    """
    Class facilitating on demand extraction of data from a RSL file.

    Parameters
    ----------
    rslfile : RslFile
        Open RslFile object to extract data from.
    volume_num : int
        Volume number of data to be extracted.
    fillvalue : int
        Value used to fill masked values in the returned array.

    """

    def __init__(self, rslfile, volume_num, fillvalue):
        """initialize the object."""
        self.rslfile = rslfile
        self.volume_num = volume_num
        self.fillvalue = fillvalue

    def __call__(self):
        """Return an array containing data from the referenced volume."""
        # extract the field and mask
        data = self.rslfile.get_volume_array(self.volume_num)
        data[np.where(np.isnan(data))] = self.fillvalue
        data[np.where(data == 131072)] = self.fillvalue
        return np.ma.masked_equal(data, self.fillvalue)


VOLUMENUM2RSLNAME = {
    0: "DZ",
    1: "VR",
    2: "SW",
    3: "CZ",
    4: "ZT",
    5: "DR",
    6: "LR",
    7: "ZD",
    8: "DM",
    9: "RH",
    10: "PH",
    11: "XZ",
    12: "CD",
    13: "MZ",
    14: "MD",
    15: "ZE",
    16: "VE",
    17: "KD",
    18: "TI",
    19: "DX",
    20: "CH",
    21: "AH",
    22: "CV",
    23: "AV",
    24: "SQ",
    25: "VS",
    26: "VL",
    27: "VG",
    28: "VT",
    29: "NP",
    30: "HC",
    31: "VC",
    32: "V2",
    33: "S2",
    34: "V3",
    35: "S3",
    36: "CR",
    37: "CC",
    38: "PR",
    39: "SD",
    40: "ZZ",
    41: "RD",
    42: "ET",
    43: "EZ",
}

RSLNAME2VOLUMENUM = {v: k for k, v in VOLUMENUM2RSLNAME.items()}



"""
Radar retrievals.

"""

from .advection import grid_displacement_pc, grid_shift  # noqa
from .comp_z import composite_reflectivity  # noqa
from .echo_class import conv_strat_yuter  # noqa
from .echo_class import get_freq_band  # noqa
from .echo_class import hydroclass_semisupervised  # noqa
from .echo_class import steiner_conv_strat  # noqa
from .gate_id import fetch_radar_time_profile, map_profile_to_gates  # noqa
from .kdp_proc import kdp_maesaka, kdp_schneebeli, kdp_vulpiani  # noqa
from .qpe import est_rain_rate_a  # noqa
from .qpe import est_rain_rate_hydro  # noqa
from .qpe import est_rain_rate_kdp  # noqa
from .qpe import est_rain_rate_z  # noqa
from .qpe import est_rain_rate_za  # noqa
from .qpe import est_rain_rate_zkdp  # noqa
from .qpe import est_rain_rate_zpoly  # noqa
from .qvp import quasi_vertical_profile  # noqa
from .simple_moment_calculations import calculate_snr_from_reflectivity  # noqa
from .simple_moment_calculations import calculate_velocity_texture  # noqa
from .simple_moment_calculations import compute_cdr  # noqa
from .simple_moment_calculations import compute_l  # noqa
from .simple_moment_calculations import compute_noisedBZ  # noqa
from .simple_moment_calculations import compute_snr  # noqa
from .spectra_calculations import dealias_spectra, spectra_moments  # noqa
from .vad import vad_browning, vad_michelson  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]
